/*
    ================================================================================
    N-PUZZLE BOARD GAME
    Copyright (c) Jarmo Pietiläinen (a.k.a. z0b) 2008-2013
    http://z0b.kapsi.fi
    --------------------------------------------------------------------------------
    This program is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License version 2.

    This program is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this program; if not, write to the Free Software Foundation, Inc., 51
    Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
    ================================================================================
*/

// Minimalistic hyperlink control I wrote many moons ago. Original code had messages for
// changing font styles and colors and everything, but I removed them.

#undef __STRICT_ANSI__
#define UNICODE
#define _UNICODE
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <string.h>
#include <wchar.h>

#include <cstdlib>

#include "hyperlinkcontrol.h"

namespace zHyperLink
{
    // --------------------------------------------------------------
    // --------------------------------------------------------------
    // DATA

    struct hyperLinkStyle_t {
        COLORREF color;
        bool underline;
    };

    // hardcoded styles
    static const hyperLinkStyle_t
        normal = { RGB(0, 0, 255), false },
        hover = { RGB(0, 0, 255), true },
        active = { RGB(255, 0, 0), true },
        visited = { RGB(0, 0, 255), false };

    struct hyperLinkData_t {
        HWND wnd;
        bool mouseHover, mouseCaptured, clicked, trackingMouse;
        RECT hoverRect, textRect;
        HCURSOR normalCursor, hoverCursor;
        HFONT font;
        wchar_t *text;
    };

    static hyperLinkData_t *getCustomWindowData(HWND wnd)
    {
        if (!wnd)
            return NULL;

        return (hyperLinkData_t *)GetWindowLongPtr(wnd, 0);
    }

    static void setCustomWindowData(HWND wnd, hyperLinkData_t *h)
    {
        if (!wnd)
            return;

        SetWindowLongPtr(wnd, 0, (LONG_PTR)h);
    }

    // --------------------------------------------------------------
    // --------------------------------------------------------------
    // WINDOW PROCEDURE

    // handles the hover state and cursor changing
    static void updateHover(HWND wnd, hyperLinkData_t *hd)
    {
        if (!wnd || !hd)
            return;

        POINT p;

        GetCursorPos(&p);
        ScreenToClient(wnd, &p);

        const bool old = hd->mouseHover;

        // detect hover
        if (PtInRect(&hd->hoverRect, p))
            hd->mouseHover = true;
        else hd->mouseHover = false;

        // detect state changes and set the cursor
        if (hd->mouseHover != old) {
            if (!hd->mouseHover)
                SetCursor(hd->normalCursor);
            else SetCursor(hd->hoverCursor);

            InvalidateRect(wnd, NULL, FALSE);
        }
    }

    // hyperlink window procedure
    static LRESULT CALLBACK hyperlinkProc(HWND wnd, UINT msg, WPARAM wparam, LPARAM lparam)
    {
        hyperLinkData_t *hd = getCustomWindowData(wnd);

        switch (msg) {
            case WM_NCCREATE:
            {
                // allocate custom structure for holding the control data
                hd = (hyperLinkData_t *)malloc(sizeof(hyperLinkData_t));

                if (!hd)
                    return FALSE;

                CREATESTRUCT *cs = (CREATESTRUCT *)lparam;

                // set the initial parameters
                hd->wnd = wnd;
                hd->font = (HFONT)GetStockObject(DEFAULT_GUI_FONT);
                hd->mouseHover = hd->mouseCaptured = hd->clicked = false;
                hd->normalCursor = LoadCursor(NULL, IDC_ARROW);
                hd->hoverCursor = LoadCursor(NULL, IDC_HAND);
                hd->text = wcsdup(cs->lpszName);
                hd->trackingMouse = false;

                // attach data to window
                setCustomWindowData(wnd, hd);

                return TRUE;
            }

            case WM_NCDESTROY:
                if (hd->text) {
                    free(hd->text);
                    hd->text = NULL;
                }

                if (hd) {
                    free(hd);
                    hd = NULL;
                }

                break;

            case WM_SETTEXT:
                if (hd->text) {
                    free(hd->text);
                    hd->text = NULL;
                }

                hd->text = wcsdup((wchar_t *)lparam);
                InvalidateRect(wnd, NULL, FALSE);

                return TRUE;

            case WM_SETFONT:
                hd->font = (HFONT)wparam;

                if (LOWORD(lparam) == TRUE)     // instant redraw
                    InvalidateRect(wnd, NULL, FALSE);

                break;

            case WM_LBUTTONDOWN:
            {
                POINT p;

                GetCursorPos(&p);
                ScreenToClient(wnd, &p);

                if (PtInRect(&hd->hoverRect, p)) {
                    hd->mouseCaptured = true;
                    SetCapture(wnd);
                    InvalidateRect(wnd, NULL, FALSE);
                }

                return 0;
            }

            case WM_LBUTTONUP:
                if (hd->mouseCaptured) {
                    hd->mouseCaptured = false;
                    ReleaseCapture();

                    POINT p;

                    GetCursorPos(&p);

                    if (WindowFromPoint(p )== wnd) {
                        ScreenToClient(wnd, &p);

                        if (PtInRect(&hd->hoverRect, p)) {
                            // we have a click!
                            hd->clicked = true;
                            hd->mouseHover = false;
                            InvalidateRect(wnd, NULL, FALSE);
                            SendMessage(GetParent(wnd), WM_COMMAND, (WPARAM)HCMD_CLICK, (LPARAM)wnd);
                        }
                    }

                    updateHover(wnd, hd);
                }

                return 0;

            // posted by TrackMouseEvent() called in WM_MOUSEMOVE
            case WM_MOUSELEAVE:
                hd->trackingMouse = false;
                updateHover(wnd, hd);
                return 0;

            case WM_MOUSEMOVE:
                updateHover(wnd, hd);

                // setup mouse leave tracking
                if (!hd->trackingMouse) {
                    TRACKMOUSEEVENT tme = { sizeof(TRACKMOUSEEVENT) };

                    tme.dwFlags = TME_LEAVE;
                    tme.hwndTrack = wnd;

                    if (TrackMouseEvent(&tme))
                        hd->trackingMouse = true;
                }

                return 0;

            case WM_ERASEBKGND:
                // we do our own buffering
                return 1;

            case WM_PAINT:
            {
                PAINTSTRUCT ps;
                RECT r;

                GetClientRect(wnd, &r);
                HDC dc = BeginPaint(wnd, &ps);

                FillRect(dc, &r, (HBRUSH)GetClassLong(wnd, GCL_HBRBACKGROUND));

                if (!hd) {
                    EndPaint(wnd, &ps);
                    return 0;
                }

                // determine the style
                const hyperLinkStyle_t *style = hd->clicked ? &visited : &normal;

                if (hd->mouseHover && !hd->mouseCaptured)
                    style = &hover;

               if (hd->mouseCaptured && hd->mouseHover)
                    style = &active;

                // draw the control
                if (hd->text) {
                    HGDIOBJ old = SelectObject(dc, hd->font);
                    COLORREF oldColor = SetTextColor(dc, style->color);
                    SIZE sz;

                    // determine the size of the text
                    SetBkMode(dc, TRANSPARENT);
                    GetTextExtentPoint32(dc, hd->text, wcslen(hd->text), &sz);

                    const int x = 1, y = 1;

                    ExtTextOut(dc, x, y, 0, &r, hd->text, wcslen(hd->text), 0);
                    SelectObject(dc, old);

                    // draw the underline
                    if (style->underline) {
                        HPEN pen = CreatePen(PS_SOLID, 0, style->color);

                        old = SelectObject(dc, pen);

                        MoveToEx(dc, x, y + sz.cy, NULL);
                        LineTo(dc, x + sz.cx, y + sz.cy);

                        SelectObject(dc, old);
                        DeleteObject(pen);
                    }

                    // Now that we know the size of the text, we can compute
                    // the exact hovering rectangle; WM_PAINT *WILL* get called
                    // before mouse can hover it...
                    SetRect(&hd->textRect, x, y, x + sz.cx, y + sz.cy + style->underline);
                    CopyRect(&hd->hoverRect, &hd->textRect);

                    // draw the focus rectangle
                    if (hd->mouseHover && hd->mouseCaptured) {
                        RECT fr;

                        CopyRect(&fr, &hd->hoverRect);
                        InflateRect(&fr, 1, 1);

                        // must restore the old color, otherwise DrawFocusRect()
                        // will draw garbage
                        SetTextColor(dc, oldColor);
                        DrawFocusRect(dc, &fr);
                    }
                }

                EndPaint(wnd, &ps);
                return 0;
            }

            default:
                break;
        }

        return DefWindowProc(wnd, msg, wparam, lparam);
    }

    // --------------------------------------------------------------
    // --------------------------------------------------------------
    // SUPPORT FUNCTIONS

    // creates the custom hyperlink control class
    bool registerClass()
    {
        WNDCLASSEX test;

        // does the window class exist?
        if (GetClassInfoEx(GetModuleHandle(NULL), className, &test) == 0) {
            // no - create it
            WNDCLASSEX wce = { sizeof(WNDCLASSEX) };

            wce.hInstance = GetModuleHandle(NULL);
            wce.lpszClassName = className;
            wce.cbWndExtra = sizeof(hyperLinkData_t *);
            wce.lpfnWndProc = hyperlinkProc;
            wce.hbrBackground = (HBRUSH)(COLOR_BTNFACE + 1);

            if (!RegisterClassEx(&wce))
                return false;
        }

        return true;
    }
}
