/*
    ================================================================================
    N-PUZZLE BOARD GAME
    Copyright (c) Jarmo Pietiläinen (a.k.a. z0b) 2008-2013
    http://z0b.kapsi.fi
    --------------------------------------------------------------------------------
    This program is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License version 2.

    This program is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this program; if not, write to the Free Software Foundation, Inc., 51
    Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
    ================================================================================
*/

#define UNICODE
#define _UNICODE
#include <windows.h>
#include <windowsx.h>
#include <commctrl.h>
#include <shlwapi.h>

#include <string>
#include <vector>
#include <cmath>
#include <cstdio>

#include "debugconsole.h"
#include "guidefs.h"
#include "globaldefs.h"
#include "options.h"
#include "npuzzle.h"
#include "util.h"
#include "hyperlinkcontrol.h"

using namespace std;

// global window handles and instances
static HINSTANCE mainInst = NULL;
static HWND mainWin = NULL,
            puzzleWin = NULL,
            statusBar = NULL;

// full path to directory where npuzzle is installed
static wstring programDir;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ABOUT DIALOG

static const wchar_t *legalese =
L"This program is free software; you can redistribute it and/or modify it\r\n"
L"under the terms of the GNU General Public License version 2 as\r\n"
L"published by the Free Software Foundation.\r\n"
L"\r\n"
L"This program is distributed in the hope that it will be useful, but\r\n"
L"WITHOUT ANY WARRANTY; without even the implied warranty of\r\n"
L"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\r\n"
L"\r\n"
L"See file LICENSE.TXT for more details. If LICENSE.TXT cannot be\r\n"
L"found from the program's directory, you have been scammed.";

// about dialog
static BOOL CALLBACK aboutDlg(HWND wnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
    switch (msg) {
        case WM_INITDIALOG:
            SetWindowText(GetDlgItem(wnd, IDC_ABOUT_NAME), L"N-Puzzle v1.2.1");
            SetWindowText(GetDlgItem(wnd, IDC_ABOUT_AUTHOR), L"Copyright (c) Jarmo Pietiläinen 2008-2013");
            SetWindowText(GetDlgItem(wnd, IDC_ABOUT_EXE), L"Compiled: " __DATE__ " " __TIME__);
            SetWindowText(GetDlgItem(wnd, IDC_ABOUT_HOME), L"http://z0b.kapsi.fi");
            SetWindowText(GetDlgItem(wnd, IDC_ABOUT_LICENSE), legalese);
            return TRUE;

        case WM_COMMAND:
            if (wparam == IDOK)
                EndDialog(wnd, IDOK);
            else if (wparam == zHyperLink::HCMD_CLICK &&
                     (HWND)lparam == GetDlgItem(wnd, IDC_ABOUT_HOME))
                ShellExecute(NULL, L"open", L"http://z0b.kapsi.fi", NULL, NULL, SW_SHOW);

            return 0;

        case WM_CLOSE:
            EndDialog(wnd, IDCANCEL);
            return 0;

        case WM_DESTROY:
            return 0;
    }

    return 0;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// CUSTOM-SIZED BOARD

// custom board size
static uint32_t newWidth, newHeight;

static void handleCustomBoardSizeSpinner(HWND edit, const int delta)
{
    if (!edit)
        return;

    int val;

    ui::getIntFromEditBox(edit, val);

    if (delta < 0)
        val--;
    else val++;

    // impose some arbitrary size limit
    val = util::clamp(val, 3, 32);
    ui::setWindowText(edit, L"%d", val);
}

// custom board size dialog
static BOOL CALLBACK customBoardDlg(HWND wnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
    switch (msg) {
        case WM_INITDIALOG:
            ui::setWindowText(GetDlgItem(wnd, IDC_CUSTOMIZE_H), L"%u", newHeight);
            ui::setWindowText(GetDlgItem(wnd, IDC_CUSTOMIZE_W), L"%u", newWidth);
            return TRUE;

        case WM_COMMAND:
            if (wparam == IDOK) {
                ui::getUIntFromEditBox(GetDlgItem(wnd, IDC_CUSTOMIZE_H), newHeight);
                ui::getUIntFromEditBox(GetDlgItem(wnd, IDC_CUSTOMIZE_W), newWidth);
                EndDialog(wnd, IDOK);
            } else if (wparam == IDCANCEL)
                EndDialog(wnd, IDCANCEL);

            return 0;

        case WM_NOTIFY:
        {
            LPNMHDR nmhdr = (LPNMHDR)lparam;

            if (nmhdr->code == UDN_DELTAPOS) {
                const int delta = -((LPNMUPDOWN)lparam)->iDelta;

                if (nmhdr->idFrom == IDC_CUSTOMIZE_HSPIN)
                    handleCustomBoardSizeSpinner(GetDlgItem(wnd, IDC_CUSTOMIZE_H), delta);
                else if (nmhdr->idFrom == IDC_CUSTOMIZE_WSPIN)
                    handleCustomBoardSizeSpinner(GetDlgItem(wnd, IDC_CUSTOMIZE_W), delta);
            }

            return 0;
        }

        case WM_CLOSE:
            EndDialog(wnd, IDCANCEL);
            return 0;

        case WM_DESTROY:
            return 0;
    }

    return 0;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// BOARD DATA AND RENDERING

// min/max size of the 3D tile border (hard limits)
static const uint32_t minTileBorderSize = 1U,
                      maxTileBorderSize = 10U;

// these are computed before the board is rendered
static uint32_t borderWidth = 0U,
                tileSize = 0U;

// rendering data
static HDC tileDC = NULL;
static HBITMAP tileImage = NULL;
static HFONT tileFont = NULL;

// moves
static uint32_t numMoves = 0U;
static void updateMoveText();

// the puzzle data
static npuzzle puzzle;

// computes suitable tile size based on font
// (sorry about this function; it is messy and old, but I don't want to change it)
static bool computeTileSize(const uint32_t width, const uint32_t height)
{
    if (!tileFont || !tileDC)
        return false;

    // determine the size of largest tile text
    HGDIOBJ old = SelectObject(tileDC, tileFont);
    int w = -1, h = -1;

    for (uint32_t y = 0U, z = 1U; y < height; y++)
        for (uint32_t x = 0U; x < width; x++) {
            static wchar_t buf[16];
            SIZE s;

            wsprintf(buf, L"%d", z++);
            GetTextExtentPoint32(tileDC, buf, wcslen(buf), &s);

            w = max(w, (int)s.cx);
            h = max(h, (int)s.cy);
        }

    SelectObject(tileDC, old);

    // determine the max size
    tileSize = max(w, h);

    tileSize += tileSize / 2;
    tileSize += 5;

    // add tweak
    tileSize = (int)((float)tileSize * opt.sizeTweak + 0.5f);

    if (tileSize < 40)
        tileSize = 40;

    // compute the border size (yes, 14 is an arbitrary value)
    borderWidth = tileSize / 14;

    if (borderWidth < minTileBorderSize)
        borderWidth = minTileBorderSize;

    if (borderWidth > maxTileBorderSize)
        borderWidth = maxTileBorderSize;

    dprintf("computeTileSize(): size=%u  border=%u\n", tileSize, borderWidth);
    return true;
}

template<class T> inline void deleteObject(T *&x)
{
    if (x) {
        DeleteObject(x);
        x = NULL;
    }
}

template<> inline void deleteObject(HDC &x)
{
    if (x) {
        DeleteDC(x);
        x = NULL;
    }
}

// deletes GDI data
static void freeTileData()
{
    deleteObject(tileFont);
    deleteObject(tileImage);
    deleteObject(tileDC);
}

// creates Windows' GDI objects used to draw the board
static bool createTileData(const uint32_t width, const uint32_t height)
{
    freeTileData();

    HDC winDC = GetDC(puzzleWin);

    if (!winDC)
        return false;

    // create the tile DC
    tileDC = CreateCompatibleDC(winDC);

    if (!tileDC) {
        ReleaseDC(puzzleWin, winDC);
        return false;
    }

    // create the font for drawing the numbers
    tileFont = CreateFont(-MulDiv(opt.fontSize, GetDeviceCaps(tileDC, LOGPIXELSY), 72), 0, 0, 0,
                          opt.fontBold ? FW_BOLD : FW_NORMAL,
                          opt.fontItalic ? TRUE : FALSE,
                          opt.fontUnderline ? TRUE : FALSE,
                          FALSE, ANSI_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS,
                          DEFAULT_QUALITY, DEFAULT_PITCH, opt.fontName.c_str());

    if (!tileFont) {
        ReleaseDC(puzzleWin, winDC);
        return false;
    }

    // now we have the DC and the font, so get the tile size
    if (!computeTileSize(width, height)) {
        ReleaseDC(puzzleWin, winDC);
        return false;
    }

    // create a bitmap for it
    tileImage = CreateCompatibleBitmap(winDC, tileSize, tileSize);
    ReleaseDC(puzzleWin, winDC);

    if (!tileImage)
        return false;

    SelectObject(tileDC, tileImage);

    // draw the tile onto a memory bitmap
    HPEN edgeHi = CreatePen(PS_SOLID, 0, opt.tileHigh),
         edgeLow = CreatePen(PS_SOLID, 0, opt.tileLow);
    HBRUSH back = CreateSolidBrush(opt.tileBack);

    if (!edgeHi || !edgeLow || !back) {
        deleteObject(edgeHi);
        deleteObject(edgeLow);
        deleteObject(back);
        return false;
    }

    // 3D lighter edge
    HGDIOBJ old = SelectObject(tileDC, edgeHi);

    for (uint32_t i = 0U; i < borderWidth; i++) {
        // -
        MoveToEx(tileDC, i, i, NULL);
        LineTo(tileDC, tileSize - i - 1, i);

        // |
        MoveToEx(tileDC, i, i + 1, NULL);
        LineTo(tileDC, i, tileSize - i - 1);
    }

    // 3D darker edge
    SelectObject(tileDC, edgeLow);

    for (uint32_t i = 0U; i < borderWidth; i++) {
        // -
        MoveToEx(tileDC, i, tileSize - 1 - i, NULL);
        LineTo(tileDC, tileSize - i, tileSize - i - 1);

        // |
        MoveToEx(tileDC, tileSize - i - 1, i, NULL);
        LineTo(tileDC, tileSize - i - 1, tileSize - i - 1);
    }

    // fill the inside
    RECT rect;

    SetRect(&rect, borderWidth, borderWidth, tileSize - borderWidth, tileSize - borderWidth);
    FillRect(tileDC, &rect, back);

    SelectObject(tileDC, old);
    deleteObject(edgeHi);
    deleteObject(edgeLow);
    deleteObject(back);

    return true;
}

// draws one tile with a number
static void drawNumberedTile(HDC dc, const uint32_t x, const uint32_t y, const uint32_t number)
{
    if (!dc || !tileDC || !tileImage || !tileFont)
        return;

    static wchar_t buf[16];

    RECT rect =
    {
        (LONG)x,
        (LONG)y,
        (LONG)(x + tileSize),
        (LONG)(y + tileSize)
    };

    SelectObject(dc, tileFont);
    SetBkMode(dc, TRANSPARENT);
    SetTextColor(dc, opt.tileFore);
    wsprintf(buf, L"%u", number);

    // blit the background tile image and draw text
    SelectObject(tileDC, tileImage);
    BitBlt(dc, x, y, tileSize, tileSize, tileDC, 0, 0, SRCCOPY);
    DrawText(dc, buf, -1, &rect, DT_SINGLELINE | DT_CENTER | DT_VCENTER);
}

// ------------------------------------------------------------------
// ------------------------------------------------------------------
// CUSTOM PUZZLE WINDOW CLASS

static int redrawPuzzle()
{
    PAINTSTRUCT ps;

    HDC dc = BeginPaint(puzzleWin, &ps);
    const RECT &draw = ps.rcPaint;

    // fill the update region's background
    HBRUSH back = CreateSolidBrush(opt.boardBack);

    FillRect(dc, &draw, back);
    DeleteObject(back);

    // then draw the tiles (cull out rows and columns)
    const int32_t dX = tileSize + opt.tilePadX,
                  dY = tileSize + opt.tilePadY;

    int32_t yp = opt.padY1;

    for (uint32_t y = 0U; y < opt.boardHeight; y++) {
        if (draw.bottom < yp || draw.top > (yp + tileSize)) {
            yp += dY;
            continue;
        }

        int32_t xp = opt.padX1;

        for (uint32_t x = 0U; x < opt.boardWidth; x++) {
            const uint32_t n = puzzle.tileAt(x, y);

            if (n == npuzzle::EMPTY || draw.right < xp || draw.left > (xp + tileSize)) {
                xp += dX;
                continue;
            }

            drawNumberedTile(dc, xp, yp, n);
            xp += dX;
        }

        yp += dY;
    }

    EndPaint(puzzleWin, &ps);
    return 0;
}

static int handleMouseClick(LPARAM lparam)
{
    if (puzzle.isSolved()) {
        dprintf("puzzle has been solved, click ignored\n");
        return 0;
    }

    const POINT mp =
    {
        GET_X_LPARAM(lparam),
        GET_Y_LPARAM(lparam)
    };

    for (uint32_t y = 0U; y < opt.boardHeight; y++) {
        for (uint32_t x = 0U; x < opt.boardWidth; x++) {
            RECT tile;

            // tile rectangle
            tile.top = opt.padY1 + y * (tileSize + opt.tilePadY);
            tile.left = opt.padX1 + x * (tileSize + opt.tilePadX);
            tile.bottom = tile.top + tileSize;
            tile.right = tile.left + tileSize;

            if (PtInRect(&tile, mp)) {
                dprintf("Clicked on tile (%d, %d) tile=%d\n", x + 1, y + 1,
                        puzzle.tileAt(x, y));

                const uint32_t dX = tileSize + opt.tilePadX,
                               dY = tileSize + opt.tilePadY;

                uint32_t emptyX, emptyY,
                         x1, x2, y1, y2;

                puzzle.getEmptyBlock(emptyX, emptyY);

                if (puzzle.move(x, y)) {
                    // This was a valid move and the board has been updated.
                    // Now we need to figure out the update rectangle. Use the same
                    // logic that npuzzle::move() uses, and determine which rows
                    // and columns are affected, and what are the start and end points;
                    // then build up a horizontal or a vertical update rectangle that
                    // covers everything that was moved.
                    RECT upd;

                    if (y == emptyY) {
                        upd.top = opt.padY1 + y * dY;
                        upd.bottom = upd.top + tileSize;

                        // one row: from column 1 to column 2
                        if (x < emptyX) {
                            x1 = x;
                            x2 = emptyX + 1;
                        } else {
                            x1 = emptyX;
                            x2 = x + 1;
                        }

                        upd.left = opt.padX1 + x1 * dX;
                        upd.right = opt.padX1 + x2 * dX;
                    } else {
                        upd.left = opt.padX1 + x * dX;
                        upd.right = upd.left + tileSize;

                        // one column: from row 1 to row 2
                        if (y < emptyY) {
                            y1 = y;
                            y2 = emptyY + 1;
                        } else {
                            y1 = emptyY;
                            y2 = y + 1;
                        }

                        upd.top = opt.padY1 + y1 * dY;
                        upd.bottom = opt.padY1 + y2 * dY;
                    }

                    numMoves++;
                    updateMoveText();

                    InvalidateRect(puzzleWin, &upd, FALSE);
                } else dprintf(" > can't move\n");

                return 0;
            }
        }
    }

    return 0;
}

// puzzle window procedure
static LRESULT CALLBACK puzzleWinProc(HWND wnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
    switch (msg) {
        case WM_ERASEBKGND:
            // ignore this, we use dirty rectangles
            return 1;

        case WM_PAINT:
            return redrawPuzzle();

        case WM_LBUTTONDOWN:
            return handleMouseClick(lparam);

        default:
            return DefWindowProc(wnd, msg, wparam, lparam);
    }

    return 0;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// MAIN WINDOW UTILITY

static void dumpErrors(const wstring &msg, HWND win)
{
    LPVOID buf;

    FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM |
                  FORMAT_MESSAGE_IGNORE_INSERTS, NULL, GetLastError(),
                  MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPTSTR)&buf,
                  0, NULL);

    if ((wchar_t *)buf == NULL)
        MessageBox(win, (msg + L"\n<No system error message found>").c_str(),
                   L"Error", MB_ICONERROR | MB_OK);
    else MessageBox(win, (msg + L"\n" + wstring((wchar_t *)buf)).c_str(),
                    L"Error", MB_ICONERROR | MB_OK);

    LocalFree(buf);
}

// resizes and repositions the main window, sets its size to puzzle window size
static void resizeWindows()
{
    if (!mainWin || !puzzleWin)
        return;

    // get the size of the puzzle window
    RECT rect;

    rect.left = 0;
    rect.top = 0;
    rect.right = opt.padX1 + opt.padX2 + opt.boardWidth * (tileSize + opt.tilePadX) - opt.tilePadX;
    rect.bottom = opt.padY1 + opt.padY2 + opt.boardHeight * (tileSize + opt.tilePadY) - opt.tilePadY;

    AdjustWindowRectEx(&rect, GetWindowLong(puzzleWin, GWL_STYLE),
                       FALSE, GetWindowLong(puzzleWin, GWL_EXSTYLE));

    const uint32_t pw = rect.right - rect.left,
                   ph = rect.bottom - rect.top;

    // statusbar height
    RECT sb;

    GetClientRect(statusBar, &sb);
    const int32_t sbH = sb.bottom - sb.top;

    // compute the size and position of the main window
    rect.top = 0;
    rect.left = 0;
    rect.right = pw;
    rect.bottom = ph + sbH;

    AdjustWindowRectEx(&rect, GetWindowLong(mainWin, GWL_STYLE),
                       TRUE, GetWindowLong(mainWin, GWL_EXSTYLE));

    const uint32_t mw = rect.right - rect.left,
                   mh = rect.bottom - rect.top;

    rect.top = GetSystemMetrics(SM_CYSCREEN) / 2 - mh / 2;
    rect.left = GetSystemMetrics(SM_CXSCREEN) / 2 - mw / 2;
    rect.right = rect.left + mw;
    rect.bottom = rect.top + mh;

    // set the positions and sizes
    SetWindowPos(mainWin, NULL, rect.left, rect.top, mw, mh, SWP_NOZORDER);
    SetWindowPos(puzzleWin, NULL, 0, 0, pw, ph, SWP_NOZORDER);
    SetWindowPos(statusBar, NULL, 0, mw - sbH, mw, sbH, SWP_NOZORDER);
}

// menu item help texts
static const struct menuItem_t {
    int id;
    const wchar_t *help;
} menuItemHelps[] =
{
    { IDM_NEW, L"Start a new game" },
    { IDM_3X3, L"Start a new game with a 3x3 board" },
    { IDM_4X4, L"Start a new game with a 4x4 board" },
    { IDM_8X8, L"Start a new game with an 8x8 board" },
    { IDM_CUSTOM, L"Start a new game with a customized board" },
    //{ IDM_SCORES, L"Shows the highscore table" },
    { IDM_PREFS, L"Change the game options" },
    { IDM_EXIT, L"Quit the game" },
    { IDM_HELP, L"Help index" },
    { IDM_ABOUT, L"About this game" },
    { 0, NULL }
};

// sets the statusbar help text
static bool updateMenuHelp(HMENU menu, const int item)
{
    if (!menu || !statusBar)
        return false;

    for (const menuItem_t *mi = menuItemHelps; mi->help; mi++)
        if (mi->id == item) {
            SendMessage(statusBar, SB_SETTEXT, 0, (LPARAM)mi->help);
            return true;
        }

    return false;
}

// updates the "moves" text in statusbar
static void updateMoveText()
{
    if (!statusBar)
        return;

    static wchar_t text[64];

    if (puzzle.isSolved())
        wsprintf(text, L"Solved the game in %u moves", numMoves);
    else wsprintf(text, L"Moves: %u", numMoves);

    SendMessage(statusBar, SB_SETTEXT, 0 , (LPARAM)text);
}

// sets the board size radio button in menu
static void updateBoardSizeMark(HWND wnd)
{
    if (!wnd)
        return;

    HMENU m = GetMenu(wnd);

    if (!m)
        return;

    uint32_t which = 3U;

    if (opt.boardWidth == 3U && opt.boardHeight == 3U)
        which = 0U;
    else if (opt.boardWidth == 4U && opt.boardHeight == 4U)
        which = 1U;
    else if (opt.boardWidth == 8U && opt.boardHeight == 8U)
        which = 2U;

    CheckMenuRadioItem(m, IDM_3X3, IDM_CUSTOM, IDM_3X3 + which, MF_BYCOMMAND);
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// MAIN WINDOW

// cleanup at exit
static void cleanUpAll()
{
    freeTileData();
    DestroyWindow(puzzleWin);
    DestroyWindow(statusBar);
    DestroyWindow(mainWin);
    console::close();
}

// starts a new game with same-sized board
static void newGameSameSize()
{
    // if the current game has been played, ask a confirmation first
    if (numMoves && !puzzle.isSolved())
        if (MessageBox(mainWin, L"Do you really want to start a new game?",
                       L"Game in progress", MB_YESNO | MB_ICONQUESTION) == IDNO)
            return;

    // init a new game
    srand(GetTickCount());

    numMoves = 0U;
    puzzle.scramble();

    updateMoveText();
    updateBoardSizeMark(mainWin);

    InvalidateRect(puzzleWin, NULL, FALSE);
}

// changes the board size and starts a new game
static void newGameAnotherSize(const int itemID)
{
    if (numMoves && !puzzle.isSolved())
        if (MessageBox(mainWin, L"Do you really want to start a new game?",
                       L"Game in progress", MB_YESNO | MB_ICONQUESTION) == IDNO)
            return;

    // figure out the new size
    uint32_t newW, newH;

    switch (itemID) {
        case IDM_3X3:
            newW = newH = 3U;
            break;

        case IDM_4X4:
            newW = newH = 4U;
            break;

        case IDM_8X8:
            newW = newH = 8U;
            break;

        case IDM_CUSTOM:
        {
            // custom size
            newWidth = opt.boardWidth;
            newHeight = opt.boardHeight;

            const int r = DialogBox(mainInst, MAKEINTRESOURCE(IDD_CUSTOMIZE),
                                    mainWin, customBoardDlg);

            if (r == -1) {
                MessageBox(mainWin, L"Failed to display the board size customization dialog.",
                           L"Error", MB_ICONSTOP | MB_OK);
                return;
            } else if (r == IDOK) {
                newW = newWidth;
                newH = newHeight;
            } else return;

            break;
        }

        default:
            dprintf("Unknown board size menuitem ID %d\n", itemID);
            return;
    }

    // initialize new game
    srand(GetTickCount());

    if (opt.boardWidth != newW || opt.boardHeight != newH) {
        // no resize if the board size does not actually change
        dprintf("New board size is %dx%d\n", newW, newH);
        opt.boardWidth = newW;
        opt.boardHeight = newH;

        if (!createTileData(opt.boardWidth, opt.boardHeight)) {
            dumpErrors(L"Can't initialize the tile drawing data.", mainWin);
            PostQuitMessage(0);
        }

        resizeWindows();
    }

    puzzle.create(newW, newH);
    puzzle.scramble();
    numMoves = 0U;

    updateMoveText();
    updateBoardSizeMark(mainWin);

    InvalidateRect(puzzleWin, NULL, FALSE);
}

static bool wantsToQuit()
{
    if (MessageBox(mainWin, L"Do you really want to quit?", L"Game in progress",
                   MB_YESNO | MB_ICONQUESTION) == IDYES)
        return true;

    return false;
}

// main window procedure
static LRESULT CALLBACK mainWindowProc(HWND wnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
    switch (msg) {
        case WM_MENUSELECT:
            if ((HMENU)lparam == NULL && (UINT)HIWORD(wparam) == 0xFFFF) {
                // set the menu item help text
                updateMoveText();
            } else {
                if (!updateMenuHelp((HMENU)lparam, LOWORD(wparam))) {
                    // failed to set help text, so display moves instead
                    updateMoveText();
                }
            }

            return 0;

        case WM_COMMAND:
            switch (LOWORD(wparam)) {
                // new game with the same board size
                case IDM_NEW:
                    newGameSameSize();
                    break;

                // new game with a new board size
                case IDM_3X3:
                case IDM_4X4:
                case IDM_8X8:
                case IDM_CUSTOM:
                    newGameAnotherSize(LOWORD(wparam));
                    break;

                case IDM_EXIT:
                    // quit - ask confirmation first if game has been played
                    if (numMoves == 0 || puzzle.isSolved())
                        PostQuitMessage(0);
                    else if (wantsToQuit())
                        PostQuitMessage(0);

                    return 0;

                case IDM_PREFS:
                {
                    // edit the game preferences
                    const int what = options::edit(wnd);

                    // recreate fonts/colors
                    if (what & options::COLORS || what & options::FONT)
                        if (!createTileData(opt.boardWidth, opt.boardHeight)) {
                            dumpErrors(L"Can't initialize the tile drawing data.", wnd);
                            PostQuitMessage(0);
                        }

                    // then resize the board/main window
                    if (what & options::FONT || what & options::PADDING)
                        resizeWindows();

                    if (what) {
                        // the window does not always get correctly resized and I'm
                        // too lazy to find out why, so I'll just redraw everything
                        ShowWindow(wnd, SW_HIDE);
                        ShowWindow(wnd, SW_SHOW);
                    }

                    return 0;
                }

                case IDM_HELP:
                {
                    const wstring fname = programDir + L"npuzzle.html";

                    // show the help
                    if (!util::fileExists(fname))
                        MessageBox(wnd, L"The help file (npuzzle.html) was not found. Make\n"
                                        L"sure it is in the same directory with the game.",
                                        L"Help file not found", MB_ICONERROR | MB_OK);
                    else ShellExecute(wnd, L"open", fname.c_str(), NULL, NULL, SW_SHOWDEFAULT);

                    return 0;
                }

                case IDM_ABOUT:
                    DialogBox(mainInst, MAKEINTRESOURCE(IDD_ABOUT), wnd, aboutDlg);
                    return 0;
            }

            return 0;

        case WM_CLOSE:
            if (numMoves == 0 || puzzle.isSolved())
                PostQuitMessage(0);
            else {
                // game in progress - ask confirmation before quitting
                if (wantsToQuit())
                    PostQuitMessage(0);
            }

            return 0;

        case WM_DESTROY:
            return 0;

        default:
            return DefWindowProc(wnd, msg, wparam, lparam);
    }

    return 0;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// MAIN

// determines the location of the main executable
static void getProgramDirectory()
{
    static const uint32_t maxPathLen = 1024U;
    static wchar_t buf[maxPathLen + 1U];

    memset(buf, 0, sizeof(wchar_t) * (maxPathLen + 1U));
    GetModuleFileName(GetModuleHandle(NULL), buf, maxPathLen);
    PathRemoveFileSpec(buf);
    PathAddBackslash(buf);

    programDir = buf;
}

int WINAPI WinMain(HINSTANCE thisInst, HINSTANCE prevInst, LPSTR cmdLine, int windowStyle)
{
    static const wchar_t mainClassName[] = L"NPuzzleWin",
                         puzzleClassName[] = L"puzzleWindowClass";

    mainInst = thisInst;
    InitCommonControls();
    srand(GetTickCount());
    getProgramDirectory();
    zHyperLink::registerClass();

    console::open(L"Debug Messages");
    options::load(programDir);

    // register the main window class
    WNDCLASSEX mwc = { sizeof(WNDCLASSEX) };

    mwc.style = CS_DBLCLKS;
    mwc.lpfnWndProc = mainWindowProc;
    mwc.hInstance = mainInst;
    mwc.hIcon = LoadIcon(mainInst, MAKEINTRESOURCE(IDC_MAINICON));
    mwc.hIconSm = LoadIcon(mainInst, MAKEINTRESOURCE(IDC_MAINICON));
    mwc.hCursor = LoadCursor(NULL, IDC_ARROW);
    mwc.hbrBackground = NULL;
    mwc.lpszClassName = mainClassName;
    mwc.lpszMenuName = MAKEINTRESOURCE(IDM_MAINMENU);

    if (!RegisterClassEx(&mwc)) {
        dumpErrors(L"Main window class registration failed.", NULL);
        cleanUpAll();
        return 1;
    }

    // create the main window (default size initially)
    mainWin = CreateWindowEx(0, mainClassName, L"N-Puzzle",
                             WS_BORDER | WS_CAPTION | WS_MINIMIZEBOX | WS_SYSMENU | WS_POPUP,
                             CW_USEDEFAULT, CW_USEDEFAULT, 640, 480,
                             HWND_DESKTOP, NULL, mainInst, NULL);

    if (!mainWin) {
        dumpErrors(L"Main window creation failed.", NULL);
        cleanUpAll();
        return 1;
    }

    // register the puzzle window class
    WNDCLASSEX pwc = { sizeof(WNDCLASSEX) };

    pwc.hInstance = mainInst;
    pwc.lpszClassName = puzzleClassName;
    pwc.lpfnWndProc = puzzleWinProc;
    pwc.style = CS_OWNDC;
    pwc.hCursor = LoadCursor(NULL, IDC_ARROW);

    if (!RegisterClassEx(&pwc)) {
        dumpErrors(L"Puzzle window class registration failed.", mainWin);
        cleanUpAll();
        return 1;
    }

    // create the puzzle window
    puzzleWin = CreateWindowEx(WS_EX_CLIENTEDGE, puzzleClassName, L"", WS_CHILD,
                               0, 0, 0, 0, mainWin, NULL, mainInst, NULL);

    if (!puzzleWin) {
        dumpErrors(L"Puzzle window creation failed.", mainWin);
        cleanUpAll();
        return 1;
    }

    // create the statusbar
    statusBar = CreateStatusWindow(WS_CHILD | WS_VISIBLE, L"", mainWin, 0);

    if (!statusBar) {
        dumpErrors(L"Can't create the statusbar.", mainWin);
        cleanUpAll();
        return 1;
    }

    // create rendering data
    if (!createTileData(opt.boardWidth, opt.boardHeight)) {
        dumpErrors(L"Can't initialize the tile drawing data.", mainWin);
        cleanUpAll();
        return 1;
    }

    // load the accelerator keys
    HACCEL accel = LoadAccelerators(mainInst, MAKEINTRESOURCE(IDA_KEYS));

    // initial game init
    numMoves = 0U;
    puzzle.create(opt.boardWidth, opt.boardHeight);
    puzzle.scramble();

    // do initial window resize
    resizeWindows();

    // initial interface update
    updateMoveText();
    updateBoardSizeMark(mainWin);

    // show everything
    ShowWindow(puzzleWin, SW_SHOW);
    ShowWindow(mainWin, windowStyle);

    // begin the main message loop
    MSG msg;

    while (GetMessage(&msg, NULL, 0, 0)) {
        if (!TranslateAccelerator(mainWin, accel, &msg)) {
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        }
    }

    // exit
    ShowWindow(mainWin, SW_HIDE);
    options::save(programDir);
    cleanUpAll();

    return msg.wParam;
}
