/****************************************************************************************
 * ncfg v0.9.2                                                                          *
 * (c) Jarmo Pietiläinen 2012-2013                                                      *
 * http://z0b.kapsi.fi                                                                  *
 ****************************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <ctype.h>

#include "ncfg.h"

#ifdef NCFG_WINDOWS
#include <windows.h>
#endif


// Branch prediction handling. This is mostly a micro-optimization, but we can do it...
#ifdef __GNUC__
    #define LIKELY(x)       __builtin_expect((x), 1)
    #define UNLIKELY(x)     __builtin_expect((x), 0)
#else
    #define LIKELY(x)       (x)
    #define UNLIKELY(x)     (x)
#endif


// strdup() doesn't always exist
static char *cfg_dup_str(const char *s)
{
    if (!s)
        return NULL;

    char *p = (char *)malloc(sizeof(char) * (strlen(s) + 1));

    if (p)
        strcpy(p, s);

    return p;
}


// Sets the error code, row and column, then returns zero. Used widely in error
// situations to avoid repeating code. The pointer is not validated, because at
// the moment, this function is never called at all if the CFG pointer is
// not already valid.
static inline int set_error(struct cfg_t *cfg, const int error,
                            const uint32_t row, const uint32_t col)
{
    cfg->error_code = error;
    cfg->error_row = row;
    cfg->error_col = col;

    return error;
}


#if defined(CFG_WANT_SAVING) || defined(CFG_WANT_LOADING)

// ======================================================================================
// I/O SYSTEM
// ======================================================================================


// wrapper for stdio fopen()
#ifdef NCFG_WINDOWS
static int stdio_open(struct cfg_io_data_t *io, const wchar_t *name, const int mode)
#else
static int stdio_open(struct cfg_io_data_t *io, const char *name, const int mode)
#endif
{
    if (!io || !name)
        return 0;

#ifdef NCFG_WINDOWS
    // _wfopen() is non-standard, Windows-only function;
    // it's either this or CreateFile() and HANDLE
    io->handle = _wfopen(name, mode == CFG_IO_MODE_READ ? L"rb" : L"wb");
#else
    // under *nix, we use UTF-8 filenames anyway
    io->handle = fopen(name, mode == CFG_IO_MODE_READ ? "rb" : "wb");
#endif

    //wprintf(L"Opening file '%s', handle=%d\n", name, io->handle == NULL ? 0 : 1);

    if (!io->handle)
        return 0;

    return 1;
}


// wrapper for stdio fclose()
static int stdio_close(struct cfg_io_data_t *io)
{
    if (io && io->handle) {
        fclose((FILE *)io->handle);
        io->handle = NULL;

        return 1;
    }

    return 0;
}


// wrapper for stdio fread()
static size_t stdio_read(struct cfg_io_data_t *input, void *out, const size_t num)
{
    return fread(out, sizeof(uint8_t), num, (FILE *)input->handle);
}


// wrapper for stdio fwrite()
static size_t stdio_write(struct cfg_io_data_t *output, const void *in, const size_t num)
{
    return fwrite(in, sizeof(uint8_t), num, (FILE *)output->handle);
}


// returns 1 if the I/O functions for requested mode have been set up properly
static int valid_io_for(struct cfg_io_functions_t *io, const int mode)
{
    if (!io)
        return 0;

    switch (mode) {
        case CFG_IO_MODE_READ:
            return io->open && io->read && io->close;

        case CFG_IO_MODE_WRITE:
            return io->open && io->write && io->close;

        default:
            break;
    }

    return 0;
}


// internal stdio I/O functions
static struct cfg_io_functions_t cfg_internal_stdio_io =
{
    .open = stdio_open,
    .close = stdio_close,
    .read = stdio_read,
    .write = stdio_write,
};


/*
// internal memory buffer I/O functions
static struct cfg_io_functions_t cfg_internal_mem_io =
{
    .open = mem_open,
    .close = mem_close,
    .read = mem_read,
    .write = mem_write,
};
*/


/*
    Makes ncfg to use the built-in stdio functions for I/O.
*/
void cfg_io_use_stdio(struct cfg_t *cfg)
{
    if (cfg)
        cfg->io = &cfg_internal_stdio_io;
}

/*
    Makes ncfg to use the built-in memory buffer functions for I/O.
    NOT YET
*/
/*
void cfg_io_use_memory_buffer(struct cfg_t *cfg)
{
    if (cfg)
        cfg->io = &cfg_internal_memory_io;
}
*/

#endif


// ======================================================================================
// SECTIONS, ITEMS AND TREE STRUCTURE MANAGEMENT
// ======================================================================================


static void internal_delete_section(struct cfg_section_t **sect);


// Deletes an item. Does not unlink it from the structure, but
// recursively deletes the child section if it exists.
static void internal_delete_item(struct cfg_item_t **item)
{
    if (!item || !*item)
        return;

    if ((*item)->key != (*item)->static_key) {
        free((*item)->key);
        (*item)->key = NULL;
    } else (*item)->key = NULL;

    cfg_item_delete_all_values(*item);

    internal_delete_section(&((*item)->child_section));

    (*item)->prev = NULL;
    (*item)->next = NULL;

    free((*item));
    (*item) = NULL;
}


// Deletes a section. Does not unlink it from the structure, but deletes the items.
static void internal_delete_section(struct cfg_section_t **sect)
{
    if (!sect || !*sect)
        return;

    // delete the items
    if ((*sect)->head) {
        struct cfg_item_t *node = (*sect)->head;

        for (; node; ) {
            struct cfg_item_t *next = node->next;

            internal_delete_item(&node);
            node = next;
        }
    }

    (*sect)->head = NULL;
    (*sect)->tail = NULL;

    free((*sect));
    (*sect) = NULL;
}


/*
    Creates a new section.
*/
int cfg_section_new(struct cfg_section_t **sect)
{
    if (!sect || *sect)
        return CFG_ARGUMENT_ERROR;

    (*sect) = (struct cfg_section_t *)malloc(sizeof(struct cfg_section_t));

    if (!(*sect))
        return CFG_OUT_OF_MEMORY;

    (*sect)->parent_item = NULL;
    (*sect)->head = NULL;
    (*sect)->tail = NULL;

    return CFG_OK;
}


/*
    Deletes a section. All items in the section are deleted. If the items
    have child sections, those child sections are also recursively deleted.
    If the section has an owner item, the link will be broken first.
*/
int cfg_section_delete(struct cfg_section_t **sect)
{
    if (!sect || !*sect)
        return CFG_ARGUMENT_ERROR;

    // remove the section from the owner
    if ((*sect)->parent_item)
        (*sect)->parent_item->child_section = NULL;

    // then nuke it
    internal_delete_section(sect);
    return CFG_OK;
}


/*
    Creates a new item. The key cannot be empty. You can pass in the initial values,
    they are duplicated and stored into the item.
*/
int cfg_item_new(struct cfg_item_t **item,
                 const char *key,
                 const size_t initial_size,
                 const char **initial_values)
{
    if (!item || *item)
        return CFG_ARGUMENT_ERROR;

    if (!key || !*key)
        return CFG_KEY_CANNOT_BE_EMPTY;

    if (initial_size && !initial_values)
        return CFG_ARGUMENT_ERROR;

    (*item) = (struct cfg_item_t *)malloc(sizeof(struct cfg_item_t));

    if (!(*item))
        return CFG_OUT_OF_MEMORY;

    (*item)->size = 0;
    (*item)->capacity = 0;
    (*item)->value = NULL;

    (*item)->prev = NULL;
    (*item)->next = NULL;

    (*item)->parent_section = NULL;
    (*item)->child_section = NULL;

    if (strlen(key) < CFG_STATIC_KEY_LENGTH) {
        strcpy((*item)->static_key, key);
        (*item)->key = (*item)->static_key;
    } else (*item)->key = cfg_dup_str(key);

    if (!(*item)->key) {
        internal_delete_item(item);
        return CFG_OUT_OF_MEMORY;
    }

    // copy initial values into exact-sized array
    if (initial_size && initial_values) {
        (*item)->value = (char **)malloc(sizeof(char *) * initial_size);

        if (!(*item)->value) {
            internal_delete_item(item);
            return CFG_OUT_OF_MEMORY;
        }

        // NULL out the pointers first...
        for (size_t n = 0; n < initial_size; n++)
            (*item)->value[n] = NULL;

        for (size_t n = 0; n < initial_size; n++) {
            (*item)->value[n] = cfg_dup_str(initial_values[n]);

            if (!(*item)->value[n]) {
                // ...because if we have to clean up, we don't
                // want to free dangling pointers
                internal_delete_item(item);
                return CFG_OUT_OF_MEMORY;
            }
        }

        (*item)->size = initial_size;
        (*item)->capacity = initial_size;
    }

    return CFG_OK;
}


/*
    Deletes an existing item. All values are destroyed. If the item
    has a child section, it will be recursively deleted. If the item
    has been linked to a section, the link will be broken first.
*/
int cfg_item_delete(struct cfg_item_t **item)
{
    if (!item || !*item)
        return CFG_ARGUMENT_ERROR;

    if ((*item)->parent_section)
        cfg_section_remove_item(*item);

    internal_delete_item(item);
    return CFG_OK;
}


/*
    Assigns a child section to an item. Returns error if the item already has a child.
*/
int cfg_item_set_child_section(struct cfg_item_t *item, struct cfg_section_t *sect)
{
    if (!item || !sect)
        return CFG_ARGUMENT_ERROR;

    if (item->child_section)
        return CFG_INVALID_DATA;

    item->child_section = sect;
    sect->parent_item = item;

    return CFG_OK;
}


/*
    Adds an item to a section. The item must not already be part of any section.
*/
int cfg_section_add_item(struct cfg_section_t *sect, struct cfg_item_t *item)
{
    if (!sect || !item)
        return CFG_ARGUMENT_ERROR;

    if (item->parent_section)
        return CFG_INVALID_DATA;

    if (!sect->head) {
        // init
        sect->head = item;
        sect->tail = item;
        item->next = NULL;
        item->prev = NULL;
    } else {
        // append
        sect->tail->next = item;
        item->prev = sect->tail;
        item->next = NULL;
        sect->tail = item;
    }

    item->parent_section = sect;

    return CFG_OK;
}


/*
    Unlinks an item from its owner section. The item is not deleted, only removed
    from the linked list of items in the owner section. The item can be then
    added to another section, or deleted, or otherwise manipulated separately from
    the tree structure.
*/
int cfg_section_remove_item(struct cfg_item_t *item)
{
    if (!item)
        return CFG_ARGUMENT_ERROR;

    if (!item->parent_section)
        return CFG_INVALID_DATA;

    if (!item->parent_section->head || !item->parent_section->tail)
        return CFG_INVALID_DATA;

    // unlink
    if (!item->prev)
        item->parent_section->head = item->next;
    else item->prev->next = item->next;

    if (!item->next)
        item->parent_section->tail = item->prev;
    else item->next->prev = item->prev;

    item->parent_section = NULL;

    return CFG_OK;
}


/*
    Renames an existing item.
*/
int cfg_item_set_key(struct cfg_item_t *item, const char *new_key)
{
    if (!item)
        return CFG_ARGUMENT_ERROR;

    if (!new_key || !*new_key)
        return CFG_KEY_CANNOT_BE_EMPTY;

    if (strlen(new_key) < CFG_STATIC_KEY_LENGTH) {
        // use the statically allocated key name
        if (item->key != item->static_key)
            free(item->key);

        strcpy(item->static_key, new_key);
        item->key = item->static_key;
    } else {
        // use a dynamically allocated key name
        char *duplicate = cfg_dup_str(new_key);

        if (!duplicate)
            return CFG_OUT_OF_MEMORY;

        if (item->key != item->static_key)
            free(item->key);

        item->key = duplicate;
    }

    return CFG_OK;
}


/*
    Gets the item's Nth value, or default value if the item/value does not exist.
*/
const char *cfg_item_get_value(const struct cfg_item_t *item,
                               const size_t index,
                               const char *default_value)
{
    if (!item || item->size == 0 || index > item->size - 1)
        return default_value;

    return item->value[index];
}


/*
    Appends a new value at the end of a value array, resizing the array if needed.
*/
int cfg_item_append_value(struct cfg_item_t *item, const char *value)
{

/*
    Keep this small, it controls how quickly the item value arrays grow.
    Most items have only one value, so it'd be stupid to grow the arrays
    unnecessarily big (even if they're "just" pointers).
*/

#define VALUE_ARRAY_GROWTH_SIZE     2

    if (!item || !value)
        return CFG_ARGUMENT_ERROR;

    if ((item->size + 1) > item->capacity) {
        // need to grow the array
        item->value =
            (char **)realloc(item->value, sizeof(char *) * (item->capacity + VALUE_ARRAY_GROWTH_SIZE));

        if (!item->value)
            return CFG_OUT_OF_MEMORY;

        item->capacity += VALUE_ARRAY_GROWTH_SIZE;

        for (size_t n = item->size; n < item->capacity; n++)
            item->value[n] = NULL;
    }

    item->value[item->size] = cfg_dup_str(value);

    if (!item->value[item->size])
        return CFG_OUT_OF_MEMORY;

    ++item->size;
    return CFG_OK;

#undef VALUE_ARRAY_GROWTH_SIZE
}


/*
    Deletes the Nth value from the item's value array.
*/
int cfg_item_delete_value(struct cfg_item_t *item, const size_t index)
{
    if (!item || item->size == 0 || index > item->size - 1)
        return CFG_ARGUMENT_ERROR;

    free(item->value[index]);

    // shift indexes, don't touch the capacity
    if (index != item->size - 1)
        for (size_t i = index; i < item->size - 1; i++)
            item->value[i] = item->value[i + 1];

    item->value[--item->size] = NULL;
    return CFG_OK;
}


/*
    Quickly deletes all the item's values, but leaves the item otherwise intact.
*/
int cfg_item_delete_all_values(struct cfg_item_t *item)
{
    if (!item)
        return CFG_ARGUMENT_ERROR;

    for (size_t n = 0; n < item->size; n++) {
        free(item->value[n]);
        item->value[n] = NULL;
    }

    free(item->value);
    item->value = NULL;

    item->size = 0;
    item->capacity = 0;

    return CFG_OK;
}


/*
    Changes the contents of an existing value.
*/
int cfg_item_set_value(struct cfg_item_t *item, const size_t index, const char *new_value)
{
    if (!item || !new_value || item->size == 0 || index > item->size - 1)
        return CFG_ARGUMENT_ERROR;

    char *duplicate = cfg_dup_str(new_value);

    if (!duplicate)
        return CFG_OUT_OF_MEMORY;

    free(item->value[index]);
    item->value[index] = duplicate;

    return CFG_OK;
}


/*
    Creates a new item and takes its first value from an integer.
*/
int cfg_item_from_int(struct cfg_item_t **item, const char *key, const int value)
{
    if (!item || *item || !key)
        return CFG_ARGUMENT_ERROR;

    char temp[256];
    const char *p = temp;

    if (snprintf(temp, 255, "%d", value) == -1)
        return CFG_VALUE_SIZE_ERROR;

    return cfg_item_new(item, key, 1, &p);
}


/*
    Creates a new item and takes its first value from a float.
*/
int cfg_item_from_float(struct cfg_item_t **item, const char *key, const float value)
{
    if (!item || *item || !key)
        return CFG_ARGUMENT_ERROR;

    char temp[256];
    const char *p = temp;

    if (snprintf(temp, 255, "%f", value) == -1)
        return CFG_VALUE_SIZE_ERROR;

    return cfg_item_new(item, key, 1, &p);
}


/*
    Appends an integer to item's value array.
*/
int cfg_item_add_int(struct cfg_item_t *item, const int value)
{
    if (!item)
        return CFG_ARGUMENT_ERROR;

    char temp[256];

    if (snprintf(temp, 255, "%d", value) == -1)
        return CFG_VALUE_SIZE_ERROR;

    return cfg_item_append_value(item, temp);
}


/*
    Appends a float to item's value array.
*/
int cfg_item_add_float(struct cfg_item_t *item, const float value)
{
    if (!item)
        return CFG_ARGUMENT_ERROR;

    char temp[256];

    if (snprintf(temp, 255, "%f", value) == -1)
        return CFG_VALUE_SIZE_ERROR;

    return cfg_item_append_value(item, temp);
}


/*
    Finds the item by name. Returns NULL if the item does not exist or something fails.
*/
const struct cfg_item_t *cfg_section_find_item(const struct cfg_section_t *s, const char *key)
{
    if (!s || !key)
        return NULL;

    for (struct cfg_item_t *item = s->head; item; item = item->next)
        if (!strcmp(item->key, key))
            return item;

    return NULL;
}


/*
    Combines cfg_section_find_item() and cfg_item_get_value() into one handy function.
*/
const char *cfg_find_and_get_value(const struct cfg_section_t *s,
                                   const char *key,
                                   const size_t index,
                                   const char *default_value)
{
    return cfg_item_get_value(cfg_section_find_item(s, key), index, default_value);
}


/*
    Returns a pointer to the specified item. Returns NULL if something goes wrong or the
    item does not exist (and 'allow_create' is zero). If the item does not exist, but
    'allow_create' is non-zero, the item is created (along with the parent items/sections)
    and a pointer to the new item is returned.
*/
struct cfg_item_t *cfg_get_item(struct cfg_section_t *root,
                                const char *path,
                                int allow_create,
                                int *error_code)
{
// max length of a single path "segment" (ie. part separated by slashes)
#define CFG_MAXIMUM_PATH_SEGMENT_LENGTH     256

#define SET_ERROR(code)                     if (error_code) *error_code = (code);

    if (!root || !path || !*path) {
        SET_ERROR(CFG_ARGUMENT_ERROR)
        return NULL;
    }

    char part[CFG_MAXIMUM_PATH_SEGMENT_LENGTH + 1];
    const char *now = path;

    struct cfg_section_t *where = root;
    struct cfg_item_t *target = NULL;

    int error = 0;

    // skip extra whitespace and slashes at the beginning
    while (*now && (isspace(*now) || *now == '/'))
        now++;

    while (1) {
        // skip extra slashes, if they exist
        if (*now == '/') {
            while (*now && *now == '/')
                now++;
        }

        // then extract next part of the path
        const char *sep = strchr(now, '/');

        if (!sep) {
            // we use the remainder of the string at once
            strcpy(part, now);
        } else {
            // get a substring
            const size_t l = sep - now;

            if (l > CFG_MAXIMUM_PATH_SEGMENT_LENGTH - 1) {
                SET_ERROR(CFG_PATH_TOO_LONG)
                return NULL;
            }

            strncpy(part, now, l);
            part[l] = '\0';
        }

        if (LIKELY(*part)) {
            // find the item from this section
            struct cfg_item_t *node = where->head;

            target = NULL;

            for (; node && !target; node = node->next)
                if (!strcmp(node->key, part))
                    target = node;

            if (target && !sep) {
                // If we've used the entire path and found the item, then this is it.
                return target;
            } else if (!target && sep) {
                // The item does not exist, but the entire path string has not been used yet.
                // We either must create the item and recurse, or stop here.
                if (!allow_create) {
                    SET_ERROR(CFG_NOT_FOUND)
                    return NULL;
                }

                // create a parent item
                struct cfg_item_t *new_item = NULL;

                if ((error = cfg_item_new(&new_item, part, 0, NULL)) != CFG_OK) {
                    SET_ERROR(error)
                    return NULL;
                }

                // add a child item
                struct cfg_section_t *new_section = NULL;

                if ((error = cfg_section_new(&new_section)) != CFG_OK) {
                    internal_delete_item(&new_item);
                    SET_ERROR(error)
                    return NULL;
                }

                cfg_item_set_child_section(new_item, new_section);
                cfg_section_add_item(where, new_item);

                // recurse
                where = new_section;
            } else if (!target && !sep) {
                // We're at the end of the path, but the item does not exist.
                if (!allow_create) {
                    SET_ERROR(CFG_NOT_FOUND)
                    return NULL;
                }

                // create the and return it
                struct cfg_item_t *new_item = NULL;

                if ((error = cfg_item_new(&new_item, part, 0, NULL)) != CFG_OK) {
                    SET_ERROR(error)
                    return NULL;
                }

                cfg_section_add_item(where, new_item);
                return new_item;
            } else if (target && !target->child_section && sep) {
                // We have the parent item, but not the child and there's still path left.
                if (!allow_create) {
                    SET_ERROR(CFG_NOT_FOUND)
                    return NULL;
                }

                // create a new child
                struct cfg_section_t *new_section = NULL;

                if ((error = cfg_section_new(&new_section)) != CFG_OK) {
                    SET_ERROR(error)
                    return NULL;
                }

                if ((error = cfg_item_set_child_section(target, new_section)) != CFG_OK) {
                    SET_ERROR(error)
                    return NULL;
                }

                // continue our search
                where = new_section;
            } else {
                if (!target || !target->child_section) {
                    // we get here if the path was not valid
                    SET_ERROR(CFG_NOT_FOUND)
                    return NULL;
                }

                // recurse
                where = target->child_section;
            }
        }

        // reached the end?
        if (!sep)
            break;

        now = sep + 1;
    }

    // we get here if we could not grok the path, most likely it was invalid
    SET_ERROR(CFG_NOT_FOUND)
    return NULL;

#undef SET_ERROR
#undef CFG_MAXIMUM_PATH_SEGMENT_LENGTH
}


// Internal worker for cfg_for_each_item() below. Does the actual work.
// Used to get rid of the pointer checks.
static int cfg_for_each_item_worker(
    struct cfg_section_t *where, const int is_recursive,
    int (*callback)(struct cfg_item_t *item, void *params), void *params)
{
    for (struct cfg_item_t *item = where->head; item; item = item->next) {
        if (!callback(item, params))
            return CFG_CALLBACK_ABORTED;

        if (is_recursive && item->child_section)
            if (cfg_for_each_item_worker(item->child_section, is_recursive, callback, params) == CFG_CALLBACK_ABORTED)
                return CFG_CALLBACK_ABORTED;
    }

    return CFG_OK;
}


/*
    Finds all matching items in a section and applies the callback function to them.
    Can work recursively.
*/
int cfg_for_each_item(struct cfg_section_t *where, const int is_recursive,
                      int (*callback)(struct cfg_item_t *item, void *params), void *params)
{
    if (!where || !callback)
        return CFG_ARGUMENT_ERROR;

    return cfg_for_each_item_worker(where, is_recursive, callback, params);
}


// Copies a NULL-terminated string into destination and then moves the destination pointer backwards.
// Done this way, because when we go upwards in the tree, we actually go backwards, so we also have
// to assemble the string backwards.
static void reverse_append(const char *source, char **destination)
{
    const size_t len = strlen(source);

    *(destination) -= len;
    memcpy(*(destination), source, len);
}


/*
    Constructs a full path for an item. Returns NULL if failed. The string returned
    must be free()'d after you're done with it.
*/
char *cfg_item_get_full_path(const struct cfg_item_t *item)
{
    if (!item)
        return NULL;

    const size_t key_len = strlen(item->key);
    size_t total_len = key_len + 1;     // +1 for the terminating NULL

    // first figure out how long the path name will be by going upwards
    // in the tree and adding together the item names
    const struct cfg_section_t *now = item->parent_section;

    while (now && now->parent_item) {
        total_len += strlen(now->parent_item->key) + 1;     // +1 for the slash
        now = now->parent_item->parent_section;
    }

    // then allocate the path and assemble it
    char *path = (char *)malloc(sizeof(char) * total_len);

    if (!path)
        return NULL;

    char *p = path + total_len;

    // first the item...
    *--p = '\0';
    reverse_append(item->key, &p);

    now = item->parent_section;

    while (now && now->parent_item) {
        // ...then each of the parent sections, separated by slashes
        *--p = '/';
        reverse_append(now->parent_item->key, &p);
        now = now->parent_item->parent_section;
    }

    return path;
}


/*
    // Constructs a full path for a section (or, to its parent item). Returns NULL if failed,
    // the string you get must be free()'d after you're done with it.
*/
char *cfg_section_get_full_path(const struct cfg_section_t *section)
{
    if (!section || !section->parent_item)
        return NULL;

    // a section that does not have a parent item does not have a path,
    // because the parent item's name is also the section's name

    return cfg_item_get_full_path(section->parent_item);
}


// ======================================================================================
// MAIN CFG STRUCTURE HANDLING
// ======================================================================================

/*
    Creates a new empty configuration data structure. Initializes I/O to
    standard STDIO functions. The root section is automatically created.
*/
int cfg_new(struct cfg_t **cfg)
{
    if (!cfg || *cfg)
        return CFG_ARGUMENT_ERROR;

    (*cfg) = (struct cfg_t *)malloc(sizeof(struct cfg_t));

    if (!(*cfg))
        return CFG_OUT_OF_MEMORY;

    (*cfg)->root = NULL;

    const int r = cfg_section_new(&((*cfg)->root));

    if (r != CFG_OK) {
        cfg_delete(cfg);
        return r;
    }

    // reset error status
    (*cfg)->error_code = CFG_OK;
    (*cfg)->error_row = 0;
    (*cfg)->error_col = 0;

#if defined(CFG_WANT_SAVING) || defined(CFG_WANT_LOADING)
    // set I/O to defaults
    (*cfg)->user_io_data = NULL;
    cfg_io_use_stdio(*cfg);
#endif

    return CFG_OK;
}


/*
    Frees the main configuration data structure.
*/
void cfg_delete(struct cfg_t **cfg)
{
    if (!cfg || !*cfg)
        return;

    internal_delete_section(&((*cfg)->root));
    (*cfg)->root = NULL;

    (*cfg)->error_code = 0;
    (*cfg)->error_row = (*cfg)->error_col = 0;

#if defined(CFG_WANT_SAVING) || defined(CFG_WANT_LOADING)
    (*cfg)->io = NULL;
#endif

    free((*cfg));
    (*cfg) = NULL;
}


/*
    Recreates the tree root, allowing it to be used for other purposes.
*/
int cfg_reset_root(struct cfg_t *cfg)
{
    if (!cfg)
        return CFG_ARGUMENT_ERROR;

    internal_delete_section(&(cfg->root));
    return cfg_section_new(&(cfg->root));
}


#ifdef CFG_WANT_LOADING

// ======================================================================================
// FILE LOADING
// ======================================================================================


// invalid UTF-8 byte sequences are replaced with this
#define REPLACEMENT_CHARACTER           0x0000FFFD


// Read buffer size. MUST be at least 3, because the buffer is, among other things, used
// to store the first bytes for the UTF-8 BOM check and that takes three bytes.
#define CFG_READ_BUFFER_SIZE            4


// initial size of the token buffer
#define CFG_INITIAL_TOKEN_CAPACITY      256


// parser data
struct parser_data_t {
    // buffer for UTF-32 codepoints
    uint32_t buffer[CFG_READ_BUFFER_SIZE];
    size_t buffer_pos;

    // token buffer, grows if needed
    size_t token_size, token_cap;
    uint8_t *token_buf;

    // I/O functions and current I/O instance
    struct cfg_io_functions_t *io;
    struct cfg_io_data_t *in;
};


// Reads and caches one UTF-32 codepoint from the input stream.
// The codepoint is removed from the cache. Returns CFG_* codes.
static int get_codepoint(struct parser_data_t *ps, uint32_t *out)
{
    if (UNLIKELY(ps->buffer_pos)) {
        // Use our buffer. Happens only if the file didn't have a UTF-8 BOM,
        // or the parser read too far while looking for a comment.
        *out = ps->buffer[--ps->buffer_pos];
        return CFG_OK;
    }

    // read in more stuff
    uint8_t b;

    if (ps->io->read(ps->in, &b, 1) != 1)
        return CFG_NORMAL_END;

    *out = b;

    if (LIKELY(*out < 0x80))        // 7-bit ASCII
        return CFG_OK;

    if (UNLIKELY(*out < 0xC0)) {    // UTF-8 forbidden area
        *out = REPLACEMENT_CHARACTER;
        return CFG_OK;
    }

    // it's a UTF-8 byte sequence, determine how much to read
    uint32_t min, mask;
    size_t need;

    if (*out < 0xE0) {              // 2 bytes
        need = 1;
        min = 0x80;
        mask = 0x3F;
    } else if (*out < 0xF0) {       // 3 bytes
        need = 2;
        min = 0x800;
        mask = 0x1F;
    } else if (*out < 0xF8) {       // 4 bytes
        need = 3;
        min = 0x1000;
        mask = 0x07;
    } else {
        // 5- and 6-byte sequences, while possible, are illegal as per RFC3629
        *out = REPLACEMENT_CHARACTER;
        return CFG_OK;
    }

    // read and add bytes
    uint8_t bytes[4];

    if (UNLIKELY(ps->io->read(ps->in, bytes, need) != need))
        return CFG_UNEXPECTED_END;

    *out &= mask;

    switch (need) {
        case 1:
            *out = (*out << 6) | (bytes[0] & 0x3F);
            break;

        case 2:
            *out = (*out << 6) | (bytes[0] & 0x3F);
            *out = (*out << 6) | (bytes[1] & 0x3F);
            break;

        case 3:
            *out = (*out << 6) | (bytes[0] & 0x3F);
            *out = (*out << 6) | (bytes[1] & 0x3F);
            *out = (*out << 6) | (bytes[2] & 0x3F);
            break;

        case 4:
            *out = (*out << 6) | (bytes[0] & 0x3F);
            *out = (*out << 6) | (bytes[1] & 0x3F);
            *out = (*out << 6) | (bytes[2] & 0x3F);
            *out = (*out << 6) | (bytes[3] & 0x3F);
            break;
    }

    if (UNLIKELY(*out < min)) {
        // invalidate overlong sequences
        *out = REPLACEMENT_CHARACTER;
        return CFG_OK;
    }

    return CFG_OK;
}


// Ungets a codepoint, ie. pushes it back in the read buffer if there's room.
// Returns CFG_* codes.
static int unget_codepoint(struct parser_data_t *ps, const uint32_t code)
{
    if (LIKELY(ps->buffer_pos + 1 < CFG_READ_BUFFER_SIZE)) {
        ps->buffer[ps->buffer_pos++] = code;
        return CFG_OK;
    }

    // oops, we ran out of space in the buffer... now what?
    return CFG_IO_ERROR;
}


// encodes a UTF-32 codepoint to UTF-8 byte sequence
static void append_to_token(struct parser_data_t *ps, const uint32_t code)
{
    uint8_t buf[4],
            *p = buf;

    // encode
    if (LIKELY(code < 0x80))
        *p++ = code & 0xFF;
    else if (code < 0x800) {
        *p++ = (code >> 6)           | 0xC0;
        *p++ = (code & 0x3F)         | 0x80;
    } else if (code < 0x10000) {
        *p++ = (code >> 12)          | 0xE0;
        *p++ = ((code >> 6) & 0x3F)  | 0x80;
        *p++ = (code & 0x3F)         | 0x80;
    } else if (code < 0x200000) {
        *p++ = (code >> 18)          | 0xF0;
        *p++ = ((code >> 12) & 0x3F) | 0x80;
        *p++ = ((code >> 6) & 0x3F)  | 0x80;
        *p++ = (code & 0x3F)         | 0x80;
    } else {
        // too long (5 or 6 bytes)
        return;
    }

    const size_t len = p - buf;

    // store (ignore 'len' here, we always double the buffer size and
    // we store mostly 1-2 bytes, so I don't think we run into problems)
    if (UNLIKELY(ps->token_size + len > ps->token_cap)) {
        ps->token_cap *= 2;
        ps->token_buf = (uint8_t *)realloc(ps->token_buf, sizeof(uint8_t) * ps->token_cap);

        if (!ps->token_buf)
            return;
    }

    memcpy(ps->token_buf + ps->token_size, buf, len);
    ps->token_size += len;
}


// token types
#define TOKEN_UNKNOWN       -1
#define TOKEN_STRING        0
#define TOKEN_GROUP_START   1
#define TOKEN_GROUP_END     2
#define TOKEN_LINE_END      3


// converts escape codes to actual byte sequences
static int unescape(struct parser_data_t *ps, const uint32_t code)
{
    switch (code) {
        case '\"':
        case '\\':
        case ';':
        case '{':
        case '}':
            append_to_token(ps, code);
            return 1;

        case 't':
            append_to_token(ps, 0x09);
            return 1;

        case 'n':
            append_to_token(ps, 0x0A);
            return 1;

        case 'r':
            append_to_token(ps, 0x0D);
            return 1;

        default:
            break;
    }

    return 0;
}


/*
    Loads configuration data from a file, using the specified I/O routines.
*/
int cfg_load(struct cfg_t *cfg,
#ifdef NCFG_WINDOWS
             const wchar_t *name)
#else
             const char *name)
#endif
{
    if (!cfg)
        return CFG_ARGUMENT_ERROR;

    if (!name)
        return set_error(cfg, CFG_ARGUMENT_ERROR, 0, 0);

    if (!valid_io_for(cfg->io, CFG_IO_MODE_READ))
        return set_error(cfg, CFG_NO_IO_FUNCTIONS, 0, 0);

    // open the file
    struct cfg_io_data_t in = { 0 };

    in.user_io_data = cfg->user_io_data;

    if (!cfg->io->open(&in, name, CFG_IO_MODE_READ))
        return set_error(cfg, CFG_CANT_OPEN_FILE, 0, 0);

    // initialize the parser buffers
    struct parser_data_t ps;

    ps.io = cfg->io;
    ps.in = &in;

    ps.token_buf = (uint8_t *)malloc(sizeof(uint8_t) * CFG_INITIAL_TOKEN_CAPACITY);

    if (!ps.token_buf) {
        cfg->io->close(&in);
        return set_error(cfg, CFG_OUT_OF_MEMORY, 0, 0);
    }

    ps.token_cap = CFG_INITIAL_TOKEN_CAPACITY;
    ps.token_size = 0;
    ps.buffer_pos = 0;

    // skip the UTF-8 BOM if it exists
    uint8_t bom[3] = { 0 };
    int numBOM = cfg->io->read(&in, bom, 3);

    if (numBOM && (numBOM != 3 || memcmp(bom, "\xEF\xBB\xBF", 3))) {
        // no BOM found, so push the bytes in the read buffer
        while (numBOM)
            unget_codepoint(&ps, bom[--numBOM]);
    }

    uint32_t code, next,
             row = 1,
             col = 1;

    int done = 0,
        n = CFG_OK,
        previous = TOKEN_UNKNOWN,
        nesting = 0;

    // Current section and item. Current section is where the items are
    // stored and current item is where the key/values are stored.
    struct cfg_section_t *curr_section = cfg->root;
    struct cfg_item_t *curr_item = NULL;

    while (!done) {
        if ((n = get_codepoint(&ps, &code)) != CFG_OK)
            break;

        const uint32_t old_col = col;

        ++col;

        switch (code) {
            case ' ':
            case '\t':
            case 0x0D:
                // skip whitespace...
                break;


            case 0x0A:
                // ...but keep track of line numbers and end comments
                ++row;
                col = 1;
                break;


            // ----------------------------------------------------------------
            // create a child section under the current item
            case '{':
            {
                struct cfg_section_t *new_section = NULL;

                if (!curr_item || curr_item->child_section || previous != TOKEN_STRING) {
                    n = CFG_INVALID_SECTION_DEFINITION;
                    col = old_col;
                    done = 1;
                    break;
                }

                if ((n = cfg_section_new(&new_section)) != CFG_OK) {
                    done = 1;
                    break;
                }

                if ((n = cfg_item_set_child_section(curr_item, new_section)) != CFG_OK) {
                    done = 1;
                    break;
                }

                // update pointers
                curr_section = new_section;
                curr_item = NULL;

                ++nesting;
                previous = TOKEN_GROUP_START;

                break;
            }


            // ----------------------------------------------------------------
            // end a child section
            case '}':
                if (previous != TOKEN_GROUP_START &&
                    previous != TOKEN_GROUP_END &&
                    previous != TOKEN_LINE_END) {
                    n = CFG_SYNTAX_ERROR;
                    col = old_col;
                    done = 1;
                }

                if (--nesting < 0) {
                    n = CFG_UNBALANCED_NESTING_LEVEL;
                    col = old_col;
                    done = 1;
                }

                curr_item = NULL;

                // return to parent section...
                if (!curr_section->parent_item) {
                    if (curr_section == cfg->root) {
                        // ...unless we're at root level, of course
                        break;
                    } else {
                        // What? A non-root section without a parent?! This should not happen.
                        // Somewhere, something went horribly wrong.
                        n = CFG_INTERNAL_PARSER_ERROR;
                        done = 1;
                        break;
                    }
                }

                // go back up to the parent section
                curr_section = curr_section->parent_item->parent_section;

                previous = TOKEN_GROUP_END;

                break;


            // ----------------------------------------------------------------
            // end of an item (ignore excessive semicolons)
            case ';':
                curr_item = NULL;
                previous = TOKEN_LINE_END;
                break;


            // ----------------------------------------------------------------
            // we found a string or a comment
            default:
            {
                ps.token_size = 0;

                // gets set to 1 if a comment was found and successfully skipped
                int got_comment = 0;

                if (code == '/') {
                    // A comment?
                    //printf("(Found / at %u)\n", get_offset(&ps));
                    n = get_codepoint(&ps, &next);
                    ++col;

                    if (n != CFG_OK) {
                        if (n == CFG_NORMAL_END) {
                            // The stream ended with a single forward slash. This is
                            // an error, because they cannot appear alone.
                            n = CFG_UNEXPECTED_END;
                        }

                        done = 1;
                        break;
                    }

                    if (next == '/') {
                        // okay, it's a start of a C++ -style comment; read until EOL
                        //printf("(Found another / at %u)\n", get_offset(&ps));

                        while (1) {
                            n = get_codepoint(&ps, &code);

                            if (n == CFG_NORMAL_END) {
                                // We got two forward slashes and that's all we need.
                                // The file can end here without an error.
                                //printf("(Normal end encountered before EOL)\n");
                                got_comment = 1;
                            }

                            if (n != CFG_OK) {
                                done = 1;
                                break;
                            }

                            if (code == 0x0A) {
                                // gotta keep track of those pesky line numbers...
                                //printf("(Found a linebreak at %u)\n", get_offset(&ps));
                                unget_codepoint(&ps, code);
                                --col;
                                got_comment = 1;
                                break;
                            }
                        }
                    } else if (next == '*') {
                        // a start of C -style comment, read until end is found
                        int comment_nesting = 0,
                            skip = 0;

                        while (1) {
                            if (code == 0x0A) {
                                // keep track of line numbers (yes, col=2 on purpose,
                                // so we don't need more complex logic)
                                //printf("(C comment linebreak at %u)\n", get_offset(&ps));
                                ++row;
                                col = 2;
                            }

                            if (skip)
                                skip--;

                            // allow C-style comments to be nested
                            if (skip == 0) {
                                if (code == '/' && next == '*') {
                                    comment_nesting++;
                                    //printf("nesting=%d\n", comment_nesting);
                                    skip = 2;
                                } else if (code == '*' && next == '/') {
                                    // found the end
                                    //printf("(Found * and / at %u)\n", get_offset(&ps));

                                    //printf("nesting=%d\n", comment_nesting - 1);

                                    if (--comment_nesting == 0) {
                                        //printf("> end\n");
                                        got_comment = 1;
                                        break;
                                    }

                                    skip = 2;
                                }
                            }

                            code = next;
                            n = get_codepoint(&ps, &next);
                            ++col;

                            if (n != CFG_OK) {
                                //printf("(2 Didn't get CFG_OK at %u)\n", get_offset(&ps));

                                if (n == CFG_NORMAL_END) {
                                    // finish what you start
                                    n = CFG_UNEXPECTED_END;
                                }

                                col = old_col + 2;
                                done = 1;
                                break;
                            }
                        }
                    } else {
                        // nah, just a slash followed by something else, do nothing
                        //printf("(False alarm, not a comment at %u)\n", get_offset(&ps));
                        unget_codepoint(&ps, next);
                        --col;
                    }
                } else if (code == '*') {
                    // a C-style comment end?
                    //printf("(Found stray * at %u)\n", get_offset(&ps));
                    n = get_codepoint(&ps, &next);

                    if (n != CFG_OK) {
                        // we want the ending quote or a semicolon, not a stream end
                        done = 1;
                        break;
                    }

                    if (next == '/') {
                        // A C-style comment end. This is a syntax error,
                        // because we handle comment ends already.
                        //printf("(Found stray */ at %u)\n", get_offset(&ps));
                        n = CFG_STRAY_COMMENT_END;
                        col = old_col;
                        done = 1;
                    } else {
                        // nah, just an asterisk followed by something else
                        //printf("(False alarm at %u)\n", get_offset(&ps));
                        unget_codepoint(&ps, next);
                        --col;
                    }
                }

                // stop if we hit an error, or skipped a comment
                if (done || got_comment)
                    break;

                // Okay, it's gotta be a string.

                // set this, so if we hit an unexpected end, we can handle it correctly
                previous = TOKEN_STRING;

                int quoted = 0;

                while (1) {
                    if (code == '\\') {
                        n = get_codepoint(&ps, &next);
                        ++col;

                        if (n != CFG_OK) {
                            done = 1;
                            break;
                        }

                        if (!unescape(&ps, next)) {
                            // Unknown escape, pass it through. It might cause problems
                            // if the file is saved again, but we don't care.
                            append_to_token(&ps, code);
                            append_to_token(&ps, next);
                        }
                    } else {
                        if (code == '\"') {
                            if (!quoted) {
                                // start of quoted string
                                quoted = 1;

                                if (ps.token_size) {
                                    // unescaped quote inside a string is an error
                                    n = CFG_SYNTAX_ERROR;
                                    done = 1;
                                    break;
                                }
                            } else {
                                // end of quoted string, stop
                                break;
                            }
                        } else {
                            // whitespace and semicolons end unquoted strings
                            if (!quoted && (code == ';' || isspace(code))) {
                                unget_codepoint(&ps, code);
                                --col;
                                break;
                            }

                            append_to_token(&ps, code);
                        }
                    }

                    n = get_codepoint(&ps, &code);
                    ++col;

                    if (n == CFG_UNEXPECTED_END) {
                        done = 1;
                        break;
                    }

                    // We got the string, but the file ended. Not an error to us,
                    // but for the other validation code this might be.
                    if (n == CFG_NORMAL_END)
                        break;
                }

                if (done)
                    break;

                // NULL-terminate the token
                append_to_token(&ps, 0);

                const char *tok = (char *)ps.token_buf;

                if (!curr_item) {
                    // create a new item
                    if (strchr(tok, '/') != NULL || strlen(tok) == 0) {
                        n = CFG_SYNTAX_ERROR;
                        col = old_col;
                        done = 1;
                        break;
                    }

                    if ((n = cfg_item_new(&curr_item, tok, 0, NULL)) != CFG_OK) {
                        col = old_col;
                        done = 1;
                        break;
                    }

                    if ((n = cfg_section_add_item(curr_section, curr_item)) != CFG_OK) {
                        col = old_col;
                        done = 1;
                        break;
                    }

                    curr_item->row = row;
                    curr_item->col = old_col;
                } else {
                    // add a value to the current item
                    if ((n = cfg_item_append_value(curr_item, tok)) != CFG_OK) {
                        done = 1;
                        break;
                    }
                }

                //printf("%3d/%3d: %s\n", row, old_col, tok);
                break;
            }
        }
    }

    free(ps.token_buf);
    cfg->io->close(&in);

    // final error checks
    if (n != CFG_NORMAL_END)
        return set_error(cfg, n, row, col);

    if (previous == TOKEN_UNKNOWN) {
        // If we hit this, it means we didn't read anything from the file.
        // Maybe it was empty? Or we hit a parser error...
        return CFG_OK;
    }

    if (nesting != 0)
        return set_error(cfg, CFG_UNBALANCED_NESTING_LEVEL, row, col);

    if (previous != TOKEN_GROUP_END && previous != TOKEN_LINE_END)
        return set_error(cfg, CFG_UNEXPECTED_END, row, col);

    // success!
    return CFG_OK;
}

#endif


#ifdef CFG_WANT_SAVING

// ======================================================================================
// WRITING STUFF TO FILES
// ======================================================================================


// indents a line
static void write_indent(struct cfg_io_functions_t *io,
                         struct cfg_io_data_t *out,
                         const size_t size)
{
    static const char indent[5] = "    ";

    // @@@TODO: indentation size must be controllable

    for (size_t i = 0; i < size; i++)
        io->write(out, indent, 4);
}


// analyzes the input string and quotes/escapes it if needed
static void write_quoted_and_escaped_string(struct cfg_io_functions_t *io,
                                            struct cfg_io_data_t *out,
                                            const char *in)
{
    if (!in || !*in) {
        // empty string must be encoded in quotes, otherwise it disappears
        io->write(out, "\"\"", 2);
        return;
    }

    const int needs_quotes = (strpbrk(in, " ") != NULL),
              needs_escaping = (strpbrk(in, "\\;\"{}\x09\x0A\x0D") != NULL);

    if (needs_quotes)
        io->write(out, "\"", 1);

    if (needs_escaping) {
        const char *p = in;

        while (*p) {
            switch (*p) {
                case '\"':
                    io->write(out, "\\\"", 2);
                    break;

                case '\\':
                    io->write(out, "\\\\", 2);
                    break;

                case ';':
                    io->write(out, "\\;", 2);
                    break;

                case '{':
                    io->write(out, "\\{", 2);
                    break;

                case '}':
                    io->write(out, "\\}", 2);
                    break;

                case 0x09:
                    io->write(out, "\\t", 2);
                    break;

                case 0x0A:
                    io->write(out, "\\n", 2);
                    break;

                case 0x0D:
                    io->write(out, "\\r", 2);
                    break;

                default:
                    io->write(out, p, 1);
                    break;
            }

            ++p;
        }
    } else {
        // no need to escape
        io->write(out, in, strlen(in));
    }

    if (needs_quotes)
        io->write(out, "\"", 1);
}


// saves one item
static int write_item(const struct cfg_item_t *item,
                      struct cfg_io_functions_t *io,
                      struct cfg_io_data_t *out,
                      const size_t indent)
{
    write_indent(io, out, indent);
    write_quoted_and_escaped_string(io, out, item->key);

    for (size_t index = 0; index < item->size; index++) {
        io->write(out, " ", 1);
        write_quoted_and_escaped_string(io, out, item->value[index]);
    }

    return CFG_OK;
}


// decide the line endings on compile-time
#ifdef NCFG_WINDOWS
    #define NL          "\x0D\x0A"
    #define LEN_NL      2
#else
    #define NL          "\x0A"
    #define LEN_NL      1
#endif


#define S_START         " {" NL
#define S_END           "}" NL
#define L_END           ";" NL


// recursively saves one section
static int write_section(const struct cfg_section_t *s,
                         struct cfg_io_functions_t *io,
                         struct cfg_io_data_t *out,
                         const size_t indent)
{
    if (!s)
        return CFG_ARGUMENT_ERROR;

    if (s->head) {
        // save the items
        const struct cfg_item_t *node = s->head;
        const size_t new_indent = indent + (!!s->parent_item);

        for (; node; ) {
            write_item(node, io, out, new_indent);

            if (node->child_section) {
                // have a child section, recurse
                io->write(out, S_START, 2 + LEN_NL);
                write_section(node->child_section, io, out, new_indent);
            } else io->write(out, L_END, 1 + LEN_NL);

            // put empty line between items without
            // childs and items with childs
            if (!node->child_section && node->next && node->next->child_section)
                io->write(out, NL, LEN_NL);

            node = node->next;
        }
    }

    if (s->parent_item) {
        write_indent(io, out, indent);
        io->write(out, S_END, 1 + LEN_NL);

        // empty line between sections
        if (s->parent_item->next)
            io->write(out, NL, LEN_NL);
    }

    return CFG_OK;
}


/*
    Saves the configuration data using the specified I/O routines.
*/
int cfg_save(struct cfg_t *cfg,
#ifdef NCFG_WINDOWS
             const wchar_t *name,
#else
             const char *name,
#endif
             const size_t num_headers,
             const char **headers)
{
    if (!cfg)
        return CFG_ARGUMENT_ERROR;

    if (!name || (num_headers && (!headers || !*headers)))
        return set_error(cfg, CFG_ARGUMENT_ERROR, 0, 0);

    if (!valid_io_for(cfg->io, CFG_IO_MODE_WRITE))
        return set_error(cfg, CFG_NO_IO_FUNCTIONS, 0, 0);

    int r;

    // if the root does not exist, create it
    if (!cfg->root)
        if ((r = cfg_section_new(&cfg->root)) != CFG_OK)
            return set_error(cfg, r, 0, 0);

    // open the file
    struct cfg_io_data_t out = { 0 };

    out.user_io_data = cfg->user_io_data;

    if (!cfg->io->open(&out, name, CFG_IO_MODE_WRITE))
        return set_error(cfg, CFG_CANT_OPEN_FILE, 0, 0);

    // write the comment header
    if (num_headers) {
        for (size_t i = 0; i < num_headers; i++) {
            cfg->io->write(&out, "// ", 3);
            cfg->io->write(&out, headers[i], strlen(headers[i]));
            cfg->io->write(&out, NL, LEN_NL);
        }

        cfg->io->write(&out, NL, LEN_NL);
    }

    // and save
    r = write_section(cfg->root, cfg->io, &out, 0);
    cfg->io->close(&out);

    if (r != CFG_OK)
        return set_error(cfg, r, 0, 0);

    return CFG_OK;
}


#undef S_START
#undef S_END
#undef L_END
#undef LEN_NL
#undef NL

#endif


// ======================================================================================
// MISCELLANEOUS
// ======================================================================================

/*
    Converts numerical error code to textual description.
*/
const char *cfg_error_code_to_string(const int code)
{
    switch (code) {
        case CFG_OK:
            return "No errors";

        case CFG_NORMAL_END:
            return "File/stream ended normally, no error";

        case CFG_UNEXPECTED_END:
            return "File/stream ended unexpectedly";

        case CFG_INVALID_UTF8:
            return "Invalid UTF-8 byte sequence";

        case CFG_IO_ERROR:
            return "Read or write operation failed";

        case CFG_UNKNOWN_ERROR:
            return "Unknown error";

        case CFG_SYNTAX_ERROR:
            return "Syntax error";

        case CFG_ARGUMENT_ERROR:
            return "Invalid function arguments";

        case CFG_INVALID_DATA:
            return "Attempted to alter the tree structure in a way that would cause problems";

        case CFG_TOO_MANY_CLOSING_BRACES:
            return "Too many }'s found";

        case CFG_UNBALANCED_NESTING_LEVEL:
            return "Nesting level not in balance";

        case CFG_STRAY_COMMENT_END:
            return "A stray C-style comment end found";

        case CFG_OUT_OF_MEMORY:
            return "Out of memory";

        case CFG_CANT_OPEN_FILE:
            return "Can't open the file or stream";

        case CFG_INTERNAL_PARSER_ERROR:
            return "The parser ended up in an unrecoverable state. Report this as a bug with a minimal testcase.";

        case CFG_INVALID_SECTION_DEFINITION:
            return "Section definition not allowed here";

        case CFG_KEY_CANNOT_BE_EMPTY:
            return "Item names cannot be empty";

        case CFG_PATH_TOO_LONG:
            return "Path name or path name segment is too long";

        case CFG_NOT_FOUND:
            return "The item does not exist";

        case CFG_NO_IO_FUNCTIONS:
            return "Required I/O functions have not been properly set up";

        case CFG_VALUE_SIZE_ERROR:
            return "String formatting failed while setting the item's value";

        case CFG_CALLBACK_ABORTED:
            return "Callback function aborted the iteration";
    }

    return "(Unknown error code passed to cfg_error_code_to_string().)";
}


const char *cfg_get_version_number(void)
{
    return "0.9.2";
}


const char *cfg_get_copyright(void)
{
    return "NCFG (c) Jarmo Pietiläinen 2012-2013";
}
