/****************************************************************************************
 * ncfg v0.9.2                                                                          *
 * (c) Jarmo Pietiläinen 2012-2013                                                      *
 * http://z0b.kapsi.fi                                                                  *
 ****************************************************************************************/

#ifndef NCFG_H
#define NCFG_H

#ifdef __cplusplus
extern "C" {
#endif

#include <string.h>
#include <stdint.h>

// detect compile-time operating system
#if defined(_WIN32) || defined(WIN32) || defined(__CYGWIN__) || defined(__MINGW32__) || defined(__BORLANDC__)
    #define NCFG_WINDOWS
    #include <wchar.h>
#endif

// size & feature optimization macros
#define CFG_WANT_LOADING
#define CFG_WANT_SAVING


// --------------------------------------------------------------------------------------
// RETURN VALUES

// See the manual for detailed explanations of what each of these mean.

enum {
    CFG_OK,
    CFG_NORMAL_END,
    CFG_UNEXPECTED_END,
    CFG_INVALID_UTF8,
    CFG_IO_ERROR,
    CFG_UNKNOWN_ERROR,
    CFG_SYNTAX_ERROR,
    CFG_INVALID_DATA,
    CFG_ARGUMENT_ERROR,
    CFG_TOO_MANY_CLOSING_BRACES,
    CFG_UNBALANCED_NESTING_LEVEL,
    CFG_STRAY_COMMENT_END,
    CFG_OUT_OF_MEMORY,
    CFG_CANT_OPEN_FILE,
    CFG_INTERNAL_PARSER_ERROR,
    CFG_INVALID_SECTION_DEFINITION,
    CFG_KEY_CANNOT_BE_EMPTY,
    CFG_PATH_TOO_LONG,
    CFG_NOT_FOUND,
    CFG_NO_IO_FUNCTIONS,
    CFG_VALUE_SIZE_ERROR,
    CFG_CALLBACK_ABORTED,
};


// --------------------------------------------------------------------------------------
// I/O

#if defined(CFG_WANT_SAVING) || defined(CFG_WANT_LOADING)

// I/O modes, mutually exclusive
#define CFG_IO_MODE_READ    0
#define CFG_IO_MODE_WRITE   1

// I/O data and handles
struct cfg_io_data_t {
    // the actual file handle
    void *handle;

    // opaque pointer to whatever data you want to store (copy of the user_io_data
    // pointer in struct cfg_t)
    void *user_io_data;
};

// I/O function pointers
struct cfg_io_functions_t {
    // Opens a file in the specified mode. Return 1 if successful, 0 if failed.
#ifdef NCFG_WINDOWS
    int (*open)(struct cfg_io_data_t *io, const wchar_t *name, const int mode);
#else
    int (*open)(struct cfg_io_data_t *io, const char *name, const int mode);
#endif

    // Closes the file. Returns 1 if OK, 0 if failed (nothing is done
    // if this fails, though...)
    int (*close)(struct cfg_io_data_t *io);

    // Attempts to read 'num' bytes of data, returns the number of actual bytes read.
    // Return 0 to signify error/end of file.
    size_t (*read)(struct cfg_io_data_t *io, void *out, const size_t num);

    // Attempts to write 'num' bytes of data, returns the number of actual bytes written.
    // Return 0 to signify error.
    size_t (*write)(struct cfg_io_data_t *io, const void *in, const size_t num);
};

#endif


// --------------------------------------------------------------------------------------
// ITEMS AND SECTIONS

struct cfg_section_t;

// to reduce excessive allocations, keys shorter than this are stored in fixed-sized arrays
#define CFG_STATIC_KEY_LENGTH   15

// an item is a key with zero or more associated values, and an optional child section
struct cfg_item_t {
    // row and column where the item was declared, valid only
    // if you load a file
    uint32_t row, col;

    // the key (always access 'key'!)
    char static_key[CFG_STATIC_KEY_LENGTH + 1];
    char *key;

    // the associated values, can be empty (dynamic array)
    size_t size, capacity;
    char **value;

    // parent section (not NULL if you load a file, but can be momentarily
    // NULL if you construct the tree by hand)
    struct cfg_section_t *parent_section;

    // child section (NULL if none)
    struct cfg_section_t *child_section;

    // linked list pointers (see the cfg_section_t structure below)
    struct cfg_item_t *prev, *next;
};

// sections have zero or more items under them
struct cfg_section_t {
    // parent item, can be NULL (for example, root sections have no parent)
    struct cfg_item_t *parent_item;

    // doubly-linked list of items in this section
    struct cfg_item_t *head, *tail;
};

// creates a new empty section
int cfg_section_new(struct cfg_section_t **sect);

// deletes a section, the items and child sections are recursively deleted if they exist
int cfg_section_delete(struct cfg_section_t **sect);

// creates a new item, optionally with initial values; the key cannot be empty or NULL
int cfg_item_new(struct cfg_item_t **item,
                 const char *key,
                 const size_t initial_size,
                 const char **initial_values);

// deletes an item and its keys; child sections and items are also recursively deleted
int cfg_item_delete(struct cfg_item_t **item);

// assigns a child section to an item
int cfg_item_set_child_section(struct cfg_item_t *item, struct cfg_section_t *sect);

// adds the item to some section
int cfg_section_add_item(struct cfg_section_t *sect, struct cfg_item_t *item);

// unlinks the item from its owner's linked list; the item is not deleted
int cfg_section_remove_item(struct cfg_item_t *item);

// appends a value to the item
int cfg_item_append_value(struct cfg_item_t *item, const char *value);

// renames an item
int cfg_item_set_key(struct cfg_item_t *item, const char *new_key);

// removes a value from an item
int cfg_item_delete_value(struct cfg_item_t *item, const size_t index);

// deletes all of the item's values, but leaves it otherwise intact
int cfg_item_delete_all_values(struct cfg_item_t *item);

// Changes an existing value. The value must already exist.
int cfg_item_set_value(struct cfg_item_t *item, const size_t index, const char *new_value);

// creates a new item and gets its first value from an integer or a float
int cfg_item_from_int(struct cfg_item_t **item, const char *key, const int value);
int cfg_item_from_float(struct cfg_item_t **item, const char *key, const float value);

// append a value to an item
int cfg_item_add_int(struct cfg_item_t *item, const int value);
int cfg_item_add_float(struct cfg_item_t *item, const float value);

// gets the Nth value from an item, or the default value if the Nth item does not exist
// (or there was some error while accessing it)
const char *cfg_item_get_value(const struct cfg_item_t *item,
                               const size_t value_index,
                               const char *default_value);

// Finds an item by name from the current section. Does NOT check the child sections
// if they have the item, therefore constructs like "a/b" are not possible.
// Returns NULL if not found or failed.
const struct cfg_item_t *cfg_section_find_item(const struct cfg_section_t *where, const char *key);

// combines cfg_section_find_item() and cfg_item_get_value() into one handy function
const char *cfg_find_and_get_value(const struct cfg_section_t *where,
                                   const char *key,
                                   const size_t value_index,
                                   const char *default_value);

// Returns a pointer to the specified item, creates it it does not exist can you
// let it to be created. Returns NULL if fails. The 'where' does not have to be
// the tree root, it can be any arbitrary section.
struct cfg_item_t *cfg_get_item(struct cfg_section_t *where,
                                const char *path,
                                int allow_create,
                                int *error_code);

// Applies the callback function to every item in the section, optionally recursively.
int cfg_for_each_item(struct cfg_section_t *where, const int is_recursive,
                      int (*callback)(struct cfg_item_t *item, void *params), void *params);

// Constructs a full path for an item. Returns NULL if failed. The string returned
// must be free()'d after you're done with it.
char *cfg_item_get_full_path(const struct cfg_item_t *item);

// Constructs a full path for a section (or, to its parent item). Returns NULL if failed,
// the string you get must be free()'d after you're done with it.
char *cfg_section_get_full_path(const struct cfg_section_t *section);


// --------------------------------------------------------------------------------------
// MAIN STRUCTURE

// the main structure
struct cfg_t {
#if defined(CFG_WANT_SAVING) || defined(CFG_WANT_LOADING)
    // I/O functions
    struct cfg_io_functions_t *io;

    // Arbitrary data you can use, it is passed to the I/O functions. This library does
    // not touch this in any way, it is up to you to allocate/free it.
    void *user_io_data;
#endif

    // root section, usually not NULL (but it doesn't hurt to check)
    struct cfg_section_t *root;

    // error code and location, valid only after calling cfg_load()
    int error_code;
    uint32_t error_row, error_col;
};

// create/delete a configuration data instance
int cfg_new(struct cfg_t **cfg);
void cfg_delete(struct cfg_t **cfg);

// clears the root
int cfg_reset_root(struct cfg_t *cfg);

#if defined(CFG_WANT_SAVING) || defined(CFG_WANT_LOADING)
// Reset the I/O functions to the internal stdio/memory buffer functions. If you want
// your own custom functions, fill in a cfg_io_functions_t struct and set cfg->io point to it.
void cfg_io_use_stdio(struct cfg_t *cfg);
//void cfg_io_use_memory_buffer(struct cfg_t *cfg); not yet :-(
#endif

#ifdef CFG_WANT_LOADING
// Parses configuration data from a file. Returns one of the CFG_* error codes.
int cfg_load(struct cfg_t *cfg,
#ifdef NCFG_WINDOWS
             const wchar_t *name);
#else
             const char *name);
#endif
#endif

#ifdef CFG_WANT_SAVING
// Saves configuration data to a file. Returns one of the CFG_* error codes.
// The headers are optional. You can use them to insert a disclaimer or a warning
// at the top of the file.
int cfg_save(struct cfg_t *cfg,
#ifdef NCFG_WINDOWS
             const wchar_t *name,
#else
             const char *name,
#endif
             const size_t num_headers,
             const char **headers);
#endif


// --------------------------------------------------------------------------------------
// MISCELLANEOUS

// error code -> string
const char *cfg_error_code_to_string(const int code);

// returns the compiled library version number
const char *cfg_get_version_number(void);

// gets a UTF-8 encoded copyright string you can display
const char *cfg_get_copyright(void);

#ifdef __cplusplus
}
#endif

#endif
