/*
    ================================================================================
    N-PUZZLE BOARD GAME
    Copyright (c) Jarmo Pietiläinen (a.k.a. z0b) 2008-2013
    http://z0b.kapsi.fi
    --------------------------------------------------------------------------------
    This program is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License version 2.

    This program is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this program; if not, write to the Free Software Foundation, Inc., 51
    Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
    ================================================================================
*/

// The actual puzzle class. 100% platform independent code.

#include <vector>
#include <cstdlib>

#include "debugconsole.h"
#include "globaldefs.h"
#include "npuzzle.h"

using namespace std;

static uint32_t getRandom()
{
    // XORshift algorithm
    static uint32_t x = 123456789U,
                    y = 362436069U,
                    z = 521288629U,
                    w = 88675123U;
    uint32_t t;

    t = x ^ (x << 11);
    x = y;
    y = z;
    z = w;
    w = w ^ (w >> 19) ^ (t ^ (t >> 8));

    // probably not correct, but this algorithm needs to be seeded some way
    return w ^ rand();
}

npuzzle::npuzzle() : width(0U), height(0U) { }

npuzzle::~npuzzle() { }

// returns true if board is OK (it can be played)
bool npuzzle::isValid() const
{
    return (width >= 3U) && (height >= 3U) && (board.size() == width * height);
}

// swaps two board blocks, DO NOT CALL IF THE BOARD IS NOT VALID!
void npuzzle::swapBlocks(const uint32_t a, const uint32_t b)
{
    const uint32_t temp = board.at(a);

    board.at(a) = board.at(b);
    board.at(b) = temp;
}

// Creates/resizes board. Smallest possible board is 3x3.
bool npuzzle::create(const uint32_t width, const uint32_t height)
{
    if (width < 3U || height < 3U)
        return false;

    try {
        this->width = width;
        this->height = height;
        board.resize(width * height);

        reset();
        return true;
    } catch (...) { }

    return false;
}

// resets the board to it's initial ("solved") state
void npuzzle::reset()
{
    if (!isValid())
        return;

    for (uint32_t i = 0U; i < (width * height); i++)
        board[i] = i + 1U;

    // place empty tile at the lower right corner
    board[width * height - 1U] = EMPTY;
}

// scrambles the board (resets it first)
void npuzzle::scramble(const uint32_t iterations)
{
    if (!isValid())
        return;

    const uint32_t numIterations = iterations ? iterations : (width * height) * 8U;

    dprintf("Scramble iterations = %u\n", numIterations);

    if (!numIterations)
        return;

    while (true) {
        reset();

        vector<pair<int, int> > moves;

        // scramble the board
        while (true) {
            for (uint32_t i = 0U; i < numIterations; i++) {
                uint32_t emptyX, emptyY;
                bool ok = false;

                getEmptyBlock(emptyX, emptyY);

                // do either horizontal or vertical move, but not both
                if (getRandom() % 10 < 5)
                    ok = move(getRandom() % width, emptyY);
                else ok = move(emptyX, getRandom() % height);

                if (ok)
                    moves.push_back(make_pair(emptyX, emptyY));
            }

            // if, for some weird reason, the board is solved after scrambling,
            // redo the scramble
            if (!isSolved())
                break;
        }

        // is the board solvable?
        vector<uint32_t> copy(board);
        bool solvable = false;

        for (vector<pair<int, int> >::reverse_iterator i = moves.rbegin();
             i != moves.rend(); ++i) {
            // if a move fails, it means board is not valid
            if (!move(i->first, i->second))
                break;

            if (isSolved()) {
                solvable = true;
                break;
            }
        }

        board = copy;

        // if solvable, stop here
        if (solvable) {
/*
            dprintf("Solve with these moves:\n");

            for (vector<pair<int, int> >::reverse_iterator i = moves.rbegin();
                 i != moves.rend(); ++i)
                dprintf("  %02d: %d, %d\n", i - moves.rbegin() + 1U, i->second + 1U, i->first + 1U);
*/

            break;
        }

        dprintf("Generated unsolvable board, retrying\n");
    }
}

// returns true if the board is in the solved state, false if not
bool npuzzle::isSolved() const
{
    if (!isValid())
        return false;

    // are the tiles in correct order?
    for (uint32_t y = 0U, z = 0U; y < height; y++)
        for (uint32_t x = 0U; x < width; x++) {
            if (board[z] != z + 1U && board[z] != EMPTY)
                return false;

            z++;
        }

    // is the lower right corner empty?
    if (board[width * height - 1U] != EMPTY)
        return false;

    return true;
}

// returns the tile at the specified coordinates
uint32_t npuzzle::tileAt(const uint32_t x, const uint32_t y) const
{
    if (!isValid())
        return INVALID;

    if ((x > width - 1U) || (y > height - 1U))
        return INVALID;

    return board[y * width + x];
}

// retrieves the position of the empty block, or (-1, -1) if board
// has not been initialized yet
void npuzzle::getEmptyBlock(uint32_t &x, uint32_t &y) const
{
    x = y = INVALID;

    if (!isValid())
        return;

    for (uint32_t yp = 0U, z = 0U; yp < height; yp++)
        for (uint32_t xp = 0U; xp < width; xp++)
            if (board[z++] == EMPTY) {
                x = xp;
                y = yp;
                return;
            }
}

// attemps a move at position (x, y), returns true if move was done (valid move)
bool npuzzle::move(const uint32_t x, const uint32_t y)
{
    if (!isValid())
        return false;

    if ((x > width - 1U) || (y > height - 1U))
        return false;

    uint32_t emptyX, emptyY;

    getEmptyBlock(emptyX, emptyY);

    // Is this a valid move? Must be either on the same row,
    // or in the same column as the empty block.
    if ((emptyX == INVALID || emptyY == INVALID) ||
        (y != emptyY && x != emptyX) ||
        (y == emptyY && x == emptyX))
        return false;

    //  It is a valid move. Move entire rows and columns.
    // (This could be simplified, but I'm lazy.)
    if (y == emptyY) {
        const uint32_t row = emptyY * width;

        if (x < emptyX) {
            for (uint32_t xp = emptyX; xp > x; xp--)
                swapBlocks(row + xp, row + xp - 1U);
        } else {
            for (uint32_t xp = emptyX; xp < x; xp++ )
                swapBlocks(row + xp, row + xp + 1U);
        }
    } else {
        if (y < emptyY) {
            for (uint32_t yp = emptyY; yp > y; yp--)
                swapBlocks(yp * width + emptyX, (yp - 1) * width + emptyX);
        } else {
            for (uint32_t yp = emptyY; yp < y; yp++)
                swapBlocks(yp * width + emptyX, (yp + 1) * width + emptyX);
        }
    }

    return true;
}
