/*
    ================================================================================
    N-PUZZLE BOARD GAME
    Copyright (c) Jarmo Pietiläinen (a.k.a. z0b) 2008-2013
    http://z0b.kapsi.fi
    --------------------------------------------------------------------------------
    This program is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License version 2.

    This program is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this program; if not, write to the Free Software Foundation, Inc., 51
    Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
    ================================================================================
*/

#define UNICODE
#define _UNICODE
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <commctrl.h>
#include <commdlg.h>
#include <cmath>
#include <string>
#include <cstdlib>
#include <cstdio>

#include <ncfg.h>

#include "guidefs.h"
#include "util.h"
#include "options.h"
#include "globaldefs.h"
#include "debugconsole.h"
#include "npuzzle.h"

using namespace std;

// default values
static const uint32_t kDefaultBoardWidth = 4U,
                      kDefaultBoardHeight = 4U,
                      kMaxBoardWidth = 32U,              // impose some limit
                      kMaxBoardHeight = 32U,
                      kMaxPadding = 64U,
                      kDefaultPadding = 16U,
                      kDefaultTilePadding = 0U,
                      kDefaultFontSize = 26U,
                      kMinFontSize = 8U,
                      kMaxFontSize = 256U;

static const float kMaxTweak = 4.0f;

// define global options
options_t opt;

// these variables store options in current dialog
static uint32_t padLeft, padRight, padTop, padBottom, padTileX, padTileY, currColor = 0U;
static float tileSizeTweak;
static COLORREF colors[5];
static wstring fontNameNow;
static uint32_t fontSizeNow;
static bool fontBoldNow, fontItalicNow, fontUnderlineNow;
static HFONT sampleFont;

// keeps track of what items changed
static int whatChanged;

static void fillLogFont(LOGFONT &f, const wstring &name,
                        const uint32_t size, const bool bold,
                        const bool italic, const bool underline,
                        HDC dc)
{
    memset(&f, 0, sizeof(LOGFONT));

    f.lfHeight = -MulDiv(size, GetDeviceCaps(dc, LOGPIXELSY), 72);
    f.lfItalic = italic ? TRUE : FALSE;
    f.lfUnderline = underline ? TRUE : FALSE;
    f.lfWeight = bold ? FW_BOLD : FW_NORMAL;
    f.lfCharSet = DEFAULT_CHARSET;
    f.lfOutPrecision = OUT_TT_ONLY_PRECIS;
    f.lfClipPrecision = CLIP_DEFAULT_PRECIS;
    f.lfQuality = DEFAULT_QUALITY;
    f.lfPitchAndFamily = DEFAULT_PITCH | FF_DONTCARE;
    wcscpy(f.lfFaceName, name.c_str());
}

// shows the font chooser
static bool chooseFont(HWND owner, HDC dc,
                       wstring &name, uint32_t &size,
                       bool &bold, bool &italic, bool &underline)
{
    LOGFONT f;

    fillLogFont(f, name, size, bold, italic, underline, dc);

    CHOOSEFONT cf = { sizeof(CHOOSEFONT) };

    cf.hwndOwner = owner;
    cf.Flags = CF_TTONLY | CF_SCREENFONTS | CF_INITTOLOGFONTSTRUCT;
    cf.lpLogFont = &f;

    if (ChooseFont(&cf)) {
        f.lfHeight = -MulDiv(f.lfHeight, 72, GetDeviceCaps(dc, LOGPIXELSY));
        name = f.lfFaceName;
        size = f.lfHeight;
        bold = (f.lfWeight == FW_BOLD) ? true : false;
        italic = f.lfItalic ? true : false;
        underline = f.lfUnderline ? true : false;

        return true;
    }

    return false;
}

// color presets
// @@@ FIXME: these should be configurable
struct preset_t {
    const wchar_t *name;
    COLORREF back, tileH, tileL, tileB, tileF;
};

static const preset_t presets[] =
{
    {
        L"Original",
        RGB(58, 110, 165), RGB(255, 255, 255), RGB(128, 128, 128), RGB(192, 192, 192), 0
    },

    {
        L"Blue",
        RGB(0, 60, 110), RGB(0, 91, 160), RGB(0, 68, 134), RGB(0, 76, 139), RGB(240, 240, 240)
    },

    {
        L"Blue Lighter",
        RGB(0, 74, 130), RGB(0, 102, 180), RGB(0, 83, 154), RGB(0, 91, 160), RGB(240, 240, 240)
    },

    {
        L"Ocean",
        RGB(0, 63, 83), RGB(0, 96, 132), RGB(0, 74, 100), RGB(0, 81, 110), RGB(227, 245, 255)
    },

    {
        L"Pastel green",
        RGB(31, 66, 31), RGB(43, 97, 43), RGB(35, 77, 36), RGB(38, 85, 38), RGB(240, 240, 240)
    },

    {
        L"Grayscale",
        RGB(59, 59, 59), RGB(89, 89, 89), RGB(68, 68, 68), RGB(75, 75, 75), RGB(240, 240, 240)
    },

    {
        L"Wine red",
        RGB(107, 24, 27), RGB(172, 35, 37), RGB(132, 29, 31), RGB(144, 30, 32), RGB(240, 240, 240)
    },

    {
        L"Magenta",
        RGB(106, 0, 106), RGB(160, 0, 161), RGB(125, 0, 127), RGB(136, 0, 137), RGB(255, 234, 255)
    },


/*
    {
        "",
        RGB(, , ), RGB(, , ), RGB(, , ), RGB(, , ), RGB(, , )
    },
*/

    { NULL }
};

static void addMenuItem(HMENU menu, const wchar_t *title, const uint32_t id)
{
    if (!menu || !title)
        return;

    MENUITEMINFO mii = { sizeof(MENUITEMINFO) };

    mii.fMask = MIIM_STATE | MIIM_TYPE | MIIM_ID;
    mii.fState = MFS_ENABLED;
    mii.fType = MFT_STRING;
    mii.wID = id;
    mii.dwTypeData = (WCHAR *)title;
    mii.cch = wcslen(title);

    InsertMenuItem(menu, 0, FALSE, &mii);
}

static void handlePadSpinner(HWND box, const int delta)
{
    if (!box)
        return;

    int val;

    ui::getIntFromEditBox(box, val);

    if (delta < 0)
        val--;
    else val++;

    val = util::clamp(val, 0, (int)kMaxPadding);
    ui::setWindowText(box, L"%u", (unsigned)val);
}

static void chooseColorForItem(HWND wnd, COLORREF &color)
{
    if (!wnd)
        return;

    static COLORREF custColors[16] =
    {
        RGB(0, 0, 0),
        RGB(0, 0, 168),
        RGB(0, 168, 0),
        RGB(0, 168, 168),
        RGB(168, 0, 0),
        RGB(168, 0, 168),
        RGB(168, 168, 0),
        RGB(168, 168, 168),
        RGB(84, 84, 84),
        RGB(84, 84, 252),
        RGB(84, 252, 84),
        RGB(84, 252, 252),
        RGB(252, 84, 84),
        RGB(252, 84, 252),
        RGB(252, 252, 84),
        RGB(252, 252, 252)
    };

    CHOOSECOLOR cc = { sizeof(CHOOSECOLOR) };

    cc.hwndOwner = wnd;
    cc.rgbResult = color;
    cc.lpCustColors = custColors;
    cc.Flags = CC_ANYCOLOR | CC_FULLOPEN | CC_RGBINIT;

    if (ChooseColor(&cc)) {
        color = cc.rgbResult;
        InvalidateRect(GetDlgItem(wnd, IDC_PREFS_COLORSAMPLE), NULL, FALSE);
    }
}

static void displayColorPresetsMenu(HWND wnd)
{
    if (!wnd)
        return;

    // create menu dynamically
    HMENU menu = CreatePopupMenu();

    if (!menu)
        return;

    int n = 1;

    for (const preset_t *p = presets; p->name; p++)
        addMenuItem(menu, p->name, n++);

    // compute menu position
    RECT r;

    GetClientRect(GetDlgItem(wnd, IDC_PREFS_PRESETS), &r);

    POINT pos = { r.right, r.top };
    ClientToScreen(GetDlgItem(wnd, IDC_PREFS_PRESETS), &pos);

    // display it
    const int ret = TrackPopupMenu(menu, TPM_LEFTALIGN | TPM_RETURNCMD,
                                   pos.x, pos.y, 0, wnd, NULL);

    if (ret >= 1 && ret <= n) {
        // activate preset
        const preset_t &p = presets[ret - 1];

        colors[0] = p.back;
        colors[1] = p.tileH;
        colors[2] = p.tileL;
        colors[3] = p.tileB;
        colors[4] = p.tileF;

        InvalidateRect(GetDlgItem(wnd, IDC_PREFS_COLORSAMPLE), NULL, FALSE);
    }
}

static void activateColor(HWND wnd, const int n)
{
    if (!wnd)
        return;

    currColor = n;
    CheckRadioButton(wnd, IDC_PREFS_BOARD, IDC_PREFS_TILEFORE, IDC_PREFS_BOARD + currColor);
    InvalidateRect(GetDlgItem(wnd, IDC_PREFS_COLORSAMPLE), NULL, FALSE);
}

static BOOL CALLBACK npPrefsDlg(HWND wnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
    switch (msg) {
        case WM_INITDIALOG:
        {
            // make copies of current settings
            padLeft = opt.padX1;
            padRight = opt.padX2;
            padTop = opt.padY1;
            padBottom = opt.padY2;
            padTileX = opt.tilePadX;
            padTileY = opt.tilePadY;
            tileSizeTweak = opt.sizeTweak;

            fontNameNow = opt.fontName;
            fontSizeNow = opt.fontSize;
            fontBoldNow = opt.fontBold;
            fontItalicNow = opt.fontItalic;
            fontUnderlineNow = opt.fontUnderline;

            colors[0] = opt.boardBack;
            colors[1] = opt.tileHigh;
            colors[2] = opt.tileLow;
            colors[3] = opt.tileBack;
            colors[4] = opt.tileFore;

            // initialize controls
            ui::setWindowText(GetDlgItem(wnd, IDC_PREFS_TOP), L"%u", padTop);
            ui::setWindowText(GetDlgItem(wnd, IDC_PREFS_LEFT), L"%u", padLeft);
            ui::setWindowText(GetDlgItem(wnd, IDC_PREFS_RIGHT), L"%u", padRight);
            ui::setWindowText(GetDlgItem(wnd, IDC_PREFS_BOTTOM), L"%u", padBottom);
            ui::setWindowText(GetDlgItem(wnd, IDC_PREFS_HPAD), L"%u", padTileX);
            ui::setWindowText(GetDlgItem(wnd, IDC_PREFS_VPAD), L"%u", padTileY);

            ui::setWindowText(GetDlgItem(wnd, IDC_PREFS_TWEAK), L"%.2f", tileSizeTweak);
            CheckRadioButton(wnd, IDC_PREFS_BOARD, IDC_PREFS_TILEFORE, IDC_PREFS_BOARD + currColor);

            // generate the preview font
            HWND fontView = GetDlgItem(wnd, IDC_PREFS_FONTSAMPLE);
            HDC dc = GetDC(fontView);
            LOGFONT f;

            fillLogFont(f, fontNameNow, fontSizeNow, fontBoldNow, fontItalicNow, fontUnderlineNow, dc);
            sampleFont = CreateFontIndirect(&f);
            ReleaseDC(fontView, dc);

            return TRUE;
        }

        case WM_DRAWITEM:
        {
            LPDRAWITEMSTRUCT ds = (LPDRAWITEMSTRUCT)lparam;
            HDC &dc = ds->hDC;
            RECT &rc = ds->rcItem;

            if (ds->CtlID == IDC_PREFS_COLORSAMPLE) {
                // color preview
                HBRUSH b = CreateSolidBrush(colors[currColor]);

                FillRect(dc, &rc, b);
                DeleteObject(b);
            } else if (ds->CtlID == IDC_PREFS_FONTSAMPLE) {
                // draw font sample
                HGDIOBJ old = SelectObject(dc, sampleFont);

                FillRect(dc, &rc, (HBRUSH)GetStockObject(WHITE_BRUSH));
                SetBkMode(dc, TRANSPARENT);
                DrawText(dc, L"0123456789", -1, &rc, DT_SINGLELINE | DT_CENTER | DT_VCENTER);
                SelectObject(dc, old);
            }

            return TRUE;
        }

        case WM_COMMAND:
            if (wparam == IDOK) {
                // get, validate and store options
                float newTweak;
                bool ok = true;

                ui::getUIntFromEditBox(GetDlgItem(wnd, IDC_PREFS_LEFT), padLeft);
                ui::getUIntFromEditBox(GetDlgItem(wnd, IDC_PREFS_RIGHT), padRight);
                ui::getUIntFromEditBox(GetDlgItem(wnd, IDC_PREFS_TOP), padTop);
                ui::getUIntFromEditBox(GetDlgItem(wnd, IDC_PREFS_BOTTOM), padBottom);
                ui::getUIntFromEditBox(GetDlgItem(wnd, IDC_PREFS_HPAD), padTileX);
                ui::getUIntFromEditBox(GetDlgItem(wnd, IDC_PREFS_VPAD), padTileY);
                ui::getFloatFromEditBox(GetDlgItem(wnd, IDC_PREFS_TWEAK), newTweak);

                // validate values
                if (newTweak < 0.01f) {
                    MessageBox(wnd,
                               L"The tile size tweaking value must be at least 0.01.",
                               L"Invalid value", MB_ICONERROR | MB_OK);
                    ok = false;
                }

                // I'm not sure can this happen, but I'll check for it anyway
                if (ok && (fontNameNow.empty() || (fontSizeNow < 1))) {
                    MessageBox(wnd,
                               L"The font name or size is invalid. Try choosing another font and/or size.",
                               L"Invalid value", MB_ICONERROR | MB_OK);
                    ok = false;
                }

                // set by spinners, but validate anyway...
                padLeft = max(0U, padLeft);
                padRight = max(0U, padRight);
                padTop = max(0U, padTop);
                padBottom = max(0U, padBottom);
                padTileX = max(0U, padTileX);
                padTileY = max(0U, padTileY);

                if (ok) {
                    // determine changed settings
                    if (padLeft != opt.padX1)
                        whatChanged |= options::PADDING;

                    if (padRight != opt.padX2)
                        whatChanged |= options::PADDING;

                    if (padTop != opt.padY1)
                        whatChanged |= options::PADDING;

                    if (padBottom != opt.padY2)
                        whatChanged |= options::PADDING;

                    if (padTileX != opt.tilePadX)
                        whatChanged |= options::PADDING;

                    if (padTileY != opt.tilePadY)
                        whatChanged |= options::PADDING;

                    if (fabs(newTweak - opt.sizeTweak) > 0.001f)
                        whatChanged |= options::PADDING;

                    if (fontNameNow != opt.fontName)
                        whatChanged |= options::FONT;

                    if (fontSizeNow != opt.fontSize)
                        whatChanged |= options::FONT;

                    if (fontBoldNow != opt.fontBold)
                        whatChanged |= options::FONT;

                    if (fontItalicNow != opt.fontItalic)
                        whatChanged |= options::FONT;

                    if (fontUnderlineNow != opt.fontUnderline)
                        whatChanged |= options::FONT;

                    if (colors[0] != opt.boardBack)
                        whatChanged |= options::COLORS;

                    if (colors[1] != opt.tileHigh)
                        whatChanged |= options::COLORS;

                    if (colors[2] != opt.tileLow)
                        whatChanged |= options::COLORS;

                    if (colors[3] != opt.tileBack)
                        whatChanged |= options::COLORS;

                    if (colors[4] != opt.tileFore)
                        whatChanged |= options::COLORS;

                    // store new (validated) settings
                    opt.padX1 = padLeft;
                    opt.padX2 = padRight;
                    opt.padY1 = padTop;
                    opt.padY2 = padBottom;
                    opt.tilePadX = padTileX;
                    opt.tilePadY = padTileY;
                    opt.sizeTweak = newTweak;

                    opt.fontName = fontNameNow;
                    opt.fontSize = fontSizeNow;
                    opt.fontBold = fontBoldNow;
                    opt.fontItalic = fontItalicNow;
                    opt.fontUnderline = fontUnderlineNow;

                    opt.boardBack = colors[0];
                    opt.tileHigh = colors[1];
                    opt.tileLow = colors[2];
                    opt.tileBack = colors[3];
                    opt.tileFore = colors[4];

                    EndDialog(wnd, IDOK);
                }
            } else if (wparam == IDCANCEL)
                EndDialog(wnd, IDCANCEL);
            else if (wparam == IDC_PREFS_BOARD)
                activateColor(wnd, 0);
            else if (wparam == IDC_PREFS_TILEHIGH)
                activateColor(wnd, 1);
            else if (wparam == IDC_PREFS_TILELOW)
                activateColor(wnd, 2);
            else if (wparam == IDC_PREFS_TILEBACK)
                activateColor(wnd, 3);
            else if (wparam == IDC_PREFS_TILEFORE)
                activateColor(wnd, 4);
            else if (wparam == IDC_PREFS_CHOOSEFONT) {
                HWND fontView = GetDlgItem(wnd, IDC_PREFS_FONTSAMPLE);
                HDC dc = GetDC(fontView);

                if (chooseFont(wnd, dc, fontNameNow, fontSizeNow, fontBoldNow,
                               fontItalicNow, fontUnderlineNow)) {
                    // recreate font and redraw the preview box
                    if (sampleFont) {
                        DeleteObject(sampleFont);
                        sampleFont = NULL;
                    }

                    LOGFONT f;

                    fillLogFont(f, fontNameNow, fontSizeNow, fontBoldNow, fontItalicNow,
                                fontUnderlineNow, dc);
                    sampleFont = CreateFontIndirect(&f);
                    InvalidateRect(GetDlgItem(wnd, IDC_PREFS_FONTSAMPLE), NULL, FALSE);
                }

                ReleaseDC(fontView, dc);
            } else if (wparam == IDC_PREFS_CHOOSECOLOR)
                chooseColorForItem(wnd, colors[currColor]);
            else if (wparam == IDC_PREFS_RESET) {
                if (MessageBox(wnd,
                               L"Restore the original settings? They will take effect immediately.",
                               L"Confirm", MB_YESNO | MB_ICONQUESTION) == IDYES) {

                    // don't reset the board size
                    const uint32_t w = opt.boardWidth,
                                   h = opt.boardHeight;

                    options::reset();
                    opt.boardWidth = w;
                    opt.boardHeight = h;

                    whatChanged = options::COLORS | options::PADDING | options::FONT;
                    EndDialog(wnd, IDOK);
                }
            } else if (wparam == IDC_PREFS_PRESETS)
                displayColorPresetsMenu(wnd);

            return 0;

        case WM_NOTIFY:
        {
            LPNMHDR nmhdr = (LPNMHDR)lparam;

            if (nmhdr->code == UDN_DELTAPOS) {
                LPNMUPDOWN p = (LPNMUPDOWN)lparam;
                const int from = nmhdr->idFrom,
                          delta = -p->iDelta;

                static const int id[6][2] =
                {
                    { IDC_PREFS_TOPSPIN, IDC_PREFS_TOP },
                    { IDC_PREFS_LEFTSPIN, IDC_PREFS_LEFT },
                    { IDC_PREFS_RIGHTSPIN, IDC_PREFS_RIGHT },
                    { IDC_PREFS_BOTTOMSPIN, IDC_PREFS_BOTTOM },
                    { IDC_PREFS_HPADSPIN, IDC_PREFS_HPAD },
                    { IDC_PREFS_VPADSPIN, IDC_PREFS_VPAD },
                };

                for (int i = 0; i < 6; i++)
                    if (from == id[i][0]) {
                        handlePadSpinner(GetDlgItem(wnd, id[i][1]), delta);
                        return 0;
                    }

                // size tweak spin
                if (nmhdr->idFrom == IDC_PREFS_TWEAKSPIN) {
                    HWND box = GetDlgItem(wnd, IDC_PREFS_TWEAK);
                    float val;

                    ui::getFloatFromEditBox(box, val);

                    if (delta < 0)
                        val -= 0.01f;
                    else val += 0.01f;

                    ui::setWindowText(box, L"%.2f", util::clamp(val, 0.01f, kMaxTweak));
                }
            }

            return 0;
        }

        case WM_CLOSE:
            EndDialog(wnd, IDCANCEL);
            return 0;

        case WM_DESTROY:
            if (sampleFont) {
                DeleteObject(sampleFont);
                sampleFont = NULL;
            }

            return 0;
    }

    return 0;
}

// If I ever build an ncfg utility library, it'll be full of functions like these:
// (except they'll be prettier, better, and more generic than these)

static void setConfigU32(cfg_t *cfg, const char *path, const uint32_t value)
{
    if (!cfg || !cfg->root)
        return;

    cfg_item_t *item = cfg_get_item(cfg->root, path, 1, NULL);

    if (item) {
        char str[128];

        sprintf(str, "%u", value);
        cfg_item_append_value(item, str);
    }
}

static void setConfigFloat(cfg_t *cfg, const char *path, const float value)
{
    if (!cfg || !cfg->root)
        return;

    cfg_item_t *item = cfg_get_item(cfg->root, path, 1, NULL);

    if (item) {
        char str[128];

        sprintf(str, "%f", value);
        cfg_item_append_value(item, str);
    }
}

static const char *getRawConfigValue(cfg_t *cfg, const char *path, const char *def = "",
                                     const size_t index = 0)
{
    return cfg_item_get_value(cfg_get_item(cfg->root, path, 0, NULL), index, def);
}

static uint32_t getClampedU32(cfg_t *cfg,
                              const char *path, const uint32_t def,
                              const uint32_t lower, const uint32_t upper,
                              const size_t index = 0)
{
    char defStr[128];

    sprintf(defStr, "%u", def);

    const char *str = getRawConfigValue(cfg, path, defStr, index);
    const uint32_t value = (uint32_t)atol(str);

    return util::clamp(value, lower, upper);
}

static uint32_t getClampedFloat(cfg_t *cfg,
                                const char *path, const float def,
                                const float lower, const float upper,
                                const size_t index = 0)
{
    char defStr[128];

    sprintf(defStr, "%f", def);

    const char *str = getRawConfigValue(cfg, path, defStr, index);
    const float value = atof(str);

    return util::clamp(value, lower, upper);
}

static bool getBoolean(cfg_t *cfg, const char *path, const bool def, const size_t index = 0)
{
    char defStr[2];

    if (def)
        defStr[0] = '1';
    else defStr[1] = '0';

    defStr[1] = '\0';

    const char *str = getRawConfigValue(cfg, path, defStr, index);

    if (!str || !*str)
        return def;

    if (!strcmp(str, "1") || !strcmp(str, "yes") || !strcmp(str, "true"))
        return true;

    return false;
}

// ------------------------------------------------------------------
// ------------------------------------------------------------------
// PUBLIC FUNCTIONS

namespace options
{
    static const wstring configName = L"npuzzle.cfg";

    // resets all settings
    void reset()
    {
        // game options
        opt.boardWidth = kDefaultBoardWidth;
        opt.boardHeight = kDefaultBoardHeight;

        // colors ("Original" theme)
        opt.boardBack = RGB(58, 110, 165);
        opt.tileHigh = RGB(255, 255, 255);
        opt.tileLow = RGB(128, 128, 128);
        opt.tileBack = RGB(192, 192, 192);
        opt.tileFore = RGB(0, 0, 0);

        // size & pad
        opt.padX1 = kDefaultPadding;
        opt.padX2 = kDefaultPadding;
        opt.padY1 = kDefaultPadding;
        opt.padY2 = kDefaultPadding;
        opt.tilePadX = kDefaultTilePadding;
        opt.tilePadY = kDefaultTilePadding;
        opt.sizeTweak = 1.0f;

        // font
        opt.fontName = L"Arial";
        opt.fontSize = kDefaultFontSize;
        opt.fontBold = true;
        opt.fontItalic = false;
        opt.fontUnderline = false;
    }

    void loadColor(const cfg_item_t *from, COLORREF &out)
    {
        if (!from || from->size != 3)
            return;

        int32_t r, g, b;

        r = util::clamp(atoi(from->value[0]), 0, 255);
        g = util::clamp(atoi(from->value[1]), 0, 255);
        b = util::clamp(atoi(from->value[2]), 0, 255);

        out = RGB(r, g, b);
    }

    static void saveColor(cfg_t *cfg, const char *path, const COLORREF &clr)
    {
        if (!cfg || !cfg->root)
            return;

        cfg_item_t *item = cfg_get_item(cfg->root, path, 1, NULL);

        if (item) {
            char str[128];

            // three separate values, not a string
            sprintf(str, "%u", GetRValue(clr));
            cfg_item_append_value(item, str);

            sprintf(str, "%u", GetGValue(clr));
            cfg_item_append_value(item, str);

            sprintf(str, "%u", GetBValue(clr));
            cfg_item_append_value(item, str);
        }
    }

    // saves settings
    bool save(const wstring &programDir)
    {
        // skip settings saving?
        if (util::fileExists(programDir + L"dont-save-settings"))
            return true;

        // save normally
        cfg_t *cfg = NULL;

        cfg_new(&cfg);

        setConfigU32(cfg, "board/width", opt.boardWidth);
        setConfigU32(cfg, "board/height", opt.boardHeight);

        saveColor(cfg, "colors/background", opt.boardBack);
        saveColor(cfg, "colors/tile_high", opt.tileHigh);
        saveColor(cfg, "colors/tile_low", opt.tileLow);
        saveColor(cfg, "colors/tile_back", opt.tileBack);
        saveColor(cfg, "colors/tile_text", opt.tileFore);

        setConfigU32(cfg, "size/pad_left", opt.padX1);
        setConfigU32(cfg, "size/pad_right", opt.padX2);
        setConfigU32(cfg, "size/pad_top", opt.padY1);
        setConfigU32(cfg, "size/pad_bottom", opt.padY2);
        setConfigU32(cfg, "size/tile_pad_x", opt.tilePadX);
        setConfigU32(cfg, "size/tile_pad_y", opt.tilePadY);
        setConfigFloat(cfg, "size/tweak", opt.sizeTweak);

        cfg_item_append_value(cfg_get_item(cfg->root, "font/name", 1, NULL), unicode::toUTF8(opt.fontName).c_str());
        setConfigU32(cfg, "font/size", opt.fontSize);
        setConfigU32(cfg, "font/bold", opt.fontBold ? 1 : 0);
        setConfigU32(cfg, "font/italic", opt.fontItalic ? 1 : 0);
        setConfigU32(cfg, "font/underline", opt.fontUnderline ? 1 : 0);

        static const char *headers[] =
        {
            "N-Puzzle v1.2.1 configuration file"
        };

        if (cfg_save(cfg, (programDir + configName).c_str(), 1, headers) != CFG_OK) {
            cfg_delete(&cfg);
            return false;
        }

        cfg_delete(&cfg);
        return true;
    }

    // loads settings
    void load(const wstring &programDir)
    {
        reset();

        cfg_t *cfg = NULL;

        cfg_new(&cfg);

        if (cfg_load(cfg, (programDir + configName).c_str()) != CFG_OK) {
            dprintf("Failed to load the options: %s, row=%u col=%u\n",
                unicode::toUCS2(cfg_error_code_to_string(cfg->error_code)).c_str(),
                cfg->error_row, cfg->error_col);
            dprintf("Options have been reseted to defaults\n");
            cfg_delete(&cfg);
            return;
        }

        opt.boardWidth = getClampedU32(cfg, "board/width", 4, 3, kMaxBoardWidth);
        opt.boardHeight = getClampedU32(cfg, "board/height", 4, 3, kMaxBoardHeight);

        loadColor(cfg_get_item(cfg->root, "colors/background", 0, NULL), opt.boardBack);
        loadColor(cfg_get_item(cfg->root, "colors/tile_high", 0, NULL), opt.tileHigh);
        loadColor(cfg_get_item(cfg->root, "colors/tile_low", 0, NULL), opt.tileLow);
        loadColor(cfg_get_item(cfg->root, "colors/tile_back", 0, NULL), opt.tileBack);
        loadColor(cfg_get_item(cfg->root, "colors/tile_text", 0, NULL), opt.tileFore);

        opt.padX1 = getClampedU32(cfg, "size/pad_left", kDefaultPadding, 0, kMaxPadding);
        opt.padX2 = getClampedU32(cfg, "size/pad_right", kDefaultPadding, 0, kMaxPadding);
        opt.padY1 = getClampedU32(cfg, "size/pad_top", kDefaultPadding, 0, kMaxPadding);
        opt.padY2 = getClampedU32(cfg, "size/pad_bottom", kDefaultPadding, 0, kMaxPadding);
        opt.tilePadX = getClampedU32(cfg, "size/tile_pad_x", kDefaultTilePadding, 0, kMaxPadding);
        opt.tilePadY = getClampedU32(cfg, "size/tile_pad_y", kDefaultTilePadding, 0, kMaxPadding);
        opt.sizeTweak = getClampedFloat(cfg, "size/tweak", 1.0f, 0.01f, kMaxTweak);

        opt.fontName = unicode::toUCS2(getRawConfigValue(cfg, "font/name", "Arial"));
        opt.fontSize = getClampedU32(cfg, "font/size", kDefaultFontSize, kMinFontSize, kMaxFontSize);
        opt.fontBold = getBoolean(cfg, "font/bold", true);
        opt.fontItalic = getBoolean(cfg, "font/italic", false);
        opt.fontUnderline = getBoolean(cfg, "font/underline", false);

        cfg_delete(&cfg);
    }

    // edit preferences
    int edit(HWND owner)
    {
        whatChanged = 0;

        const int r = DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_PREFS),
                                owner, npPrefsDlg);

        if (r == -1) {
            MessageBox(owner, L"Failed to display the Preferences dialog.", L"Error",
                       MB_ICONSTOP | MB_OK);
            return 0;
        } else if (r == IDCANCEL)
            return 0;

        return whatChanged;
    }
}
